import os
from typing import Callable, Optional

import numpy as np
import pytest
from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.distributions import (Boolean, Choice, LogUniform,
                                                RandInt)
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import AutoML, RandomSearchSampler
from neuraxle.metaopt.callbacks import (EarlyStoppingCallback, MetricCallback,
                                        ScoringCallback)
from neuraxle.metaopt.data.vanilla import (AutoMLContext,
                                           VanillaHyperparamsRepository)
from neuraxle.metaopt.validation import ValidationSplitter
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import NumpyRavel
from neuraxle.steps.output_handlers import OutputTransformerWrapper
from neuraxle.steps.sklearn import SKLearnWrapper
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.preprocessing import StandardScaler
from testing.metaopt.test_automl_repositories import CX_WITH_REPO_CTORS, TmpDir


def _create_data_source():
    data_inputs = np.random.random((25, 50)).astype(np.float32)
    expected_outputs = (np.random.random((25,)) > 0.5).astype(np.int32)
    return data_inputs, expected_outputs


def _create_pipeline():
    return Pipeline([
        StandardScaler(),
        OutputTransformerWrapper(NumpyRavel()),
        SKLearnWrapper(
            LogisticRegression(),
            HyperparameterSpace({
                'C': LogUniform(0.01, 10.0),
                'fit_intercept': Boolean(),
                'penalty': Choice(['none', 'l2']),
                'max_iter': RandInt(20, 200)
            })
        )
    ])


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_automl_api_entry_point(tmpdir, cx_repo_ctor: Callable[[Optional[TmpDir]], AutoMLContext]):
    data_inputs, expected_outputs = _create_data_source()
    dact = DACT(data_inputs=data_inputs, expected_outputs=expected_outputs)
    pipeline = _create_pipeline()
    # TODO: # HyperbandControllerLoop(), ClusteringParallelFor() ?

    a: AutoML = AutoML(
        pipeline=pipeline,
        validation_splitter=ValidationSplitter(0.20),
        hyperparams_optimizer=RandomSearchSampler(),
        hyperparams_repository=VanillaHyperparamsRepository(cache_folder=os.path.join(tmpdir, "hp")),
        scoring_callback=ScoringCallback(mean_squared_error),
        callbacks=[
            MetricCallback('accuracy', metric_function=accuracy_score, higher_score_is_better=False),
            EarlyStoppingCallback(max_epochs_without_improvement=3)
        ],
        continue_loop_on_error=True,
        n_trials=17,
        epochs=11,
        refit_best_trial=True,
    )

    a, _out = a.handle_fit_transform(
        dact,
        cx_repo_ctor()
    )

    assert _out is not None



"""
Create label encoder across multiple columns
==================================================

You can apply label encoder to all columns using the ColumnTransformer step.

This demonstrates how to use properly transform columns using neuraxle.

For more info, see the `thread here <https://stackoverflow.com/questions/24458645/label-encoding-across-multiple-columns-in-scikit-learn>`__.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import numpy as np
import pandas
from sklearn.preprocessing import LabelEncoder

from neuraxle.steps.column_transformer import ColumnTransformer
from neuraxle.steps.loop import FlattenForEach

# Discussion:
# https://stackoverflow.com/questions/24458645/label-encoding-across-multiple-columns-in-scikit-learn
df = pandas.DataFrame({
    'pets': ['cat', 'dog', 'cat', 'monkey', 'dog', 'dog'],
    'owner': ['Champ', 'Ron', 'Brick', 'Champ', 'Veronica', 'Ron'],
    'location': ['San_Diego', 'New_York', 'New_York', 'San_Diego', 'San_Diego', 'New_York']
})


def _apply_same_encoder_to_all_columns():
    """
    One shared LabelEncoder will be applied on all the data to encode it.
    """
    p = FlattenForEach(LabelEncoder(), then_unflatten=True)

    p, predicted_output = p.fit_transform(df.values)

    expected_output = np.array([
        [6, 7, 6, 8, 7, 7],
        [1, 3, 0, 1, 5, 3],
        [4, 2, 2, 4, 4, 2]
    ]).transpose()
    assert np.array_equal(predicted_output, expected_output)


def _apply_different_encoders_to_columns():
    """
    One standalone LabelEncoder will be applied on the pets,
    and another one will be shared for the columns owner and location.
    """
    p = ColumnTransformer([
        # A different encoder will be used for column 0 with name "pets":
        (0, FlattenForEach(LabelEncoder(), then_unflatten=True)),
        # A shared encoder will be used for column 1 and 2, "owner" and "location":
        ([1, 2], FlattenForEach(LabelEncoder(), then_unflatten=True)),
    ], n_dimension=2)

    p, predicted_output = p.fit_transform(df.values)

    expected_output = np.array([
        [0, 1, 0, 2, 1, 1],
        [1, 3, 0, 1, 5, 3],
        [4, 2, 2, 4, 4, 2]
    ]).transpose()
    assert np.array_equal(predicted_output, expected_output)


def main():
    _apply_same_encoder_to_all_columns()
    _apply_different_encoders_to_columns()


if __name__ == "__main__":
    main()



import numpy as np

from neuraxle.pipeline import Pipeline
from neuraxle.steps.data import DataShuffler
from neuraxle.steps.misc import TapeCallbackFunction, FitTransformCallbackStep


def test_data_shuffling_should_shuffle_data_inputs_and_expected_outputs():
    callback_fit = TapeCallbackFunction()
    callback_transform = TapeCallbackFunction()
    data_shuffler = Pipeline([
        DataShuffler(seed=42, increment_seed_after_each_fit=True),
        FitTransformCallbackStep(callback_transform, callback_fit)
    ])
    data_inputs = np.array(range(10))
    expected_outputs = np.array(range(10, 20))

    outputs = data_shuffler.fit_transform(data_inputs, expected_outputs)

    assert not np.array_equal(outputs, data_inputs)
    assert not np.array_equal(callback_fit.data[0][0], data_inputs)
    assert not np.array_equal(callback_fit.data[0][1], expected_outputs)
    assert not np.array_equal(callback_transform.data, data_inputs)



"""
Inverse Transforms in Neuraxle: How to Reverse a Prediction
============================================================

This demonstrates how to make a prediction, and then to undo the prediction to get back the original inputs or an
estimate of the original inputs. Not every pipeline steps have an inverse transform method, because not every operation
is reversible.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import numpy as np

from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN


def main():
    p = Pipeline([MultiplyByN(multiply_by=2)])

    data_inputs = np.array([1, 2])
    generated_outputs = p.transform(data_inputs)
    regenerated_inputs = p.inverse_transform(generated_outputs)

    assert np.array_equal(regenerated_inputs, data_inputs)
    assert np.array_equal(generated_outputs, 2 * data_inputs)


if __name__ == "__main__":
    main()



"""
Neuraxle's AutoML Scope Manager Classes
====================================


Scope manager objects used by AutoML algorithm classes.

Although the name "manager" for a class is disliked by the Neuraxle community,
it is used as these are Pythonic context managers, as described in
`Python's contextlib documentation <https://docs.python.org/3/library/contextlib.html>`__.

Classes are splitted like this for the AutoML:
- Projects
- Clients
- Rounds (runs)
- Trials
- TrialSplits
- MetricResults


..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""

import copy
import gc
import hashlib
import os
import typing
from abc import abstractmethod
from collections import OrderedDict
from types import TracebackType
from typing import (Callable, ContextManager, Dict, Generic, Iterable, List,
                    Optional, Tuple, Type, TypeVar)

import numpy as np
from neuraxle.base import BaseService
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import Flow, TrialStatus, _CouldHaveContext
from neuraxle.hyperparams.space import (FlatDict, HyperparameterSamples,
                                        HyperparameterSpace, RecursiveDict)
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_PROJECT,
                                           RETRAIN_TRIAL_SPLIT_ID,
                                           AutoMLContext, BaseDataclass,
                                           BaseHyperparameterOptimizer,
                                           ClientDataclass,
                                           HyperparamsRepository,
                                           MetricResultsDataclass,
                                           ProjectDataclass, RootDataclass,
                                           RoundDataclass, ScopedLocation,
                                           ScopedLocationAttr,
                                           ScopedLocationAttrInt,
                                           SubDataclassT, TrialDataclass,
                                           TrialSplitDataclass,
                                           VanillaHyperparamsRepository,
                                           dataclass_2_id_attr)

SubAggregateT = TypeVar('SubAggregateT', bound=Optional['BaseAggregate'])
ParentAggregateT = TypeVar('ParentAggregateT', bound=Optional['BaseAggregate'])


def _with_method_as_context_manager(
    func: Callable[['BaseAggregate'], 'BaseAggregate']
) -> Callable[['BaseAggregate'], ContextManager['SubAggregateT']]:
    """
    This is a method to be used as a context manager.
    This will sync items with the repos.
    Example:

    .. code-block:: python
        with obj.func() as managed_context:
            # obj.__enter__() is called
            managed_context.do_something()
        # obj.__exit__() is called

    .. seealso::
        :class:`neuraxle.metaopt.data.aggregates.BaseAggregate`
        :func:`neuraxle.metaopt.data.aggregates.BaseAggregate.__enter__`
        :func:`neuraxle.metaopt.data.aggregates.BaseAggregate.__exit__`
        :class:`neuraxle.metaopt.auto_ml.Trainer`
        :func:`neuraxle.metaopt.auto_ml.Trainer.train`

    """
    # Also add the docstring of `_with_method_as_context_manager' to the docstring of the `func`:
    func.__doc__ = (func.__doc__ or "") + _with_method_as_context_manager.__doc__
    return func


class BaseAggregate(_CouldHaveContext, BaseService, ContextManager[SubAggregateT], Generic[ParentAggregateT, SubAggregateT, SubDataclassT]):
    """
    Base class for aggregated objects using the repo and the dataclasses to manipulate them.

    For more information, read this `article by Martin Fowler on DDD Aggregates <https://martinfowler.com/bliki/DDD_Aggregate.html>`_.
    """

    def __init__(self, _dataclass: SubDataclassT, context: AutoMLContext, is_deep=False, parent: ParentAggregateT = None):
        BaseService.__init__(self, name=f"{self.__class__.__name__}_{_dataclass.get_id()}")
        _CouldHaveContext.__init__(self)
        self._dataclass: SubDataclassT = _dataclass
        self._spare: SubDataclassT = copy.copy(_dataclass).shallow()
        # TODO: pre-push context to allow for dc auto-loading and easier parent auto-loading?
        self.context: AutoMLContext = context.push_attr(_dataclass)
        self.loc: ScopedLocation = self.context.loc.copy()
        self.is_deep = is_deep
        self._parent: ParentAggregateT = parent

        self.service_assertions = [Flow, HyperparamsRepository]
        self._invariant()

    @staticmethod
    def from_context(context: AutoMLContext, is_deep=True) -> 'BaseAggregate':
        """
        Return an aggregate from a context.
        Requirement: Have already pre-push the attr of dataclass into
        the context before calling this.
        """
        _dataclass = context.load_dc(deep=True)
        aggregate_class = dataclass_2_aggregate[_dataclass.__class__]
        return aggregate_class(_dataclass, context.pop_attr(), is_deep=is_deep)

    @classmethod
    def dummy(cls: Type['BaseAggregate'], context: CX = None) -> 'BaseAggregate':
        """
        Create a dummy object of the desired type for testing purposes.
        The parent subtree will be created in a temporary repository,
        or in a real repository if the context is provided.
        """
        context = context or AutoMLContext.from_context(context)

        # create dataclass tree:
        dataclasses: List[BaseDataclass] = dataclass_2_id_attr.keys()
        root_dc: RootDataclass = RootDataclass()
        target_dc_class: Type[BaseDataclass] = aggregate_2_dataclass[cls]
        prev_dc: BaseDataclass = root_dc
        for _dataclass in dataclasses:
            # loop for every type of dataclass and create them until we find the one we want:
            dc: BaseDataclass = _dataclass()
            prev_dc.store(dc)

            prev_dc = dc
            context = context.push_attr(dc)
            if isinstance(dc, target_dc_class):
                break

        # create the aggregate:
        context.repo.save(root_dc, ScopedLocation(), deep=True)
        return cls.from_context(context, is_deep=False)

    def _invariant(self):
        _type: Type[SubDataclassT] = self.dataclass
        self._assert(isinstance(self._dataclass, _type),
                     f"self._dataclass should be of type {_type.__name__} but is of type "
                     f"{self._dataclass.__class__.__name__}", self.context)
        self._assert(isinstance(self.context, AutoMLContext),
                     f"self.context should be of type AutoMLContext but is of type "
                     f"{self.context.__class__.__name__}", self.context)
        self._assert(isinstance(self.loc, ScopedLocation),
                     f"self.loc should be of type ScopedLocation but is of type "
                     f"{self.loc.__class__.__name__}", self.context)
        self._assert(self.loc.as_list() == self.context.loc.as_list(),
                     f"{self.loc} should be equal to {self.context.loc}", self.context)
        self._assert(self.loc[-1] == self._dataclass.get_id(),
                     f"{self.loc}'s last attr should be equal to self._dataclass.get_id() "
                     f"and the id is {self._dataclass.get_id()}", self.context)
        self._assert_at_lifecycle(self.context)

    def without_context(self) -> 'BaseAggregate':
        """
        Return a copy of this aggregate without the context.
        Useful for initializing a temporary aggregate,
        such as a filtered or reduced aggregate without all its subaggregates
        to disallow saving the reduced aggregate.
        """
        self_copy = copy.copy(self)
        self_copy.context = None
        return self_copy

    @property
    def flow(self) -> Flow:
        return self.context.flow

    @property
    def repo(self) -> HyperparamsRepository:
        return self.context.repo

    def subaggregate(self, _dataclass: SubDataclassT, context: AutoMLContext, is_deep=False, parent: ParentAggregateT = None) -> SubAggregateT:
        return aggregate_2_subaggregate[self.__class__](_dataclass, context, is_deep, parent)

    @property
    def parent(self) -> ParentAggregateT:
        if self._parent is None and not isinstance(self, Root):
            self._parent = self.from_context(self.context.pop_attr(), is_deep=False)
            self._assert(isinstance(self._parent, self.parentaggregate),
                         f"parent should be of type {self.parentaggregate.__name__} but is of type {self._parent.__class__.__name__}", self.context)
        return self._parent

    @property
    def parentaggregate(self) -> Type[ParentAggregateT]:
        return {v: k for k, v in aggregate_2_subaggregate.items()}[self.__class__]

    @property
    def dataclass(self) -> Type[BaseDataclass]:
        return aggregate_2_dataclass[self.__class__]

    @property
    def subdataclass(self) -> Type[SubDataclassT]:
        return aggregate_2_dataclass[aggregate_2_subaggregate[self.__class__]]

    def sanitize_metric_name(self, metric_name: str = None):
        """
        If the argument metric is None, the optimizer's metric is taken.
        If the optimizer's metric is None, the parent metric is taken.
        """
        if metric_name is not None:
            return metric_name
        elif hasattr(self._dataclass, "main_metric_name") and self._dataclass.main_metric_name is not None:
            return self._dataclass.main_metric_name
        return metric_name or (self.parent.sanitize_metric_name(metric_name) if self.parent is not None else None)

    def refresh(self, deep: bool = True):
        with self.context.lock:
            _new_dc: SubDataclassT = self.repo.load(self.loc, deep=deep)

            def _fail_on_unsaved_changes(_self_now: SubDataclassT, _self_before: SubDataclassT, _self_new_loaded: SubDataclassT):
                _has_self_changed = _self_now != _self_before
                if _has_self_changed:
                    _is_new_dc_different_from_now = _self_now != _self_new_loaded
                    if _is_new_dc_different_from_now:

                        raise RuntimeError(f"{_self_now.__class__.__name__} with id `{_self_now.get_id()}` "
                                           f"has unsaved changes while it tried to refresh itself. "
                                           f"Before changes: {_self_before}. "
                                           f"After changes, that is not saved: {_self_now}. "
                                           f"DC that we tried to load: {_self_new_loaded}. ")
            _fail_on_unsaved_changes(
                self._dataclass.shallow(),
                self._spare.shallow(),
                _new_dc.shallow()
            )

            self._dataclass = _new_dc

        self.is_deep = deep
        self._invariant()

    def save(self, deep: bool = True) -> 'BaseAggregate':
        if deep and deep != self.is_deep:
            raise ValueError(
                f"Cannot save {str(self)} with deep=True when self "
                f"is not already deep. You might want to use self.refresh(deep=True) at "
                f"some point to refresh self before saving deeply then.")

        self._invariant()
        self._spare = copy.copy(self._dataclass).shallow()

        with self.context.lock:
            self.repo.save(self._dataclass, self.loc, deep=deep)
        return self

    def save_subaggregate(self, subagg: SubAggregateT, deep=False) -> 'BaseAggregate':
        self._dataclass.store(subagg._dataclass)

        with self.context.lock:
            self.save(deep=False)
            subagg.save(deep=deep)
        return self

    def __enter__(self) -> SubAggregateT:
        # self.context.free_scoped_logger_handler_file()
        self._invariant()
        self._managed_resource._invariant()

        self._managed_resource.context.add_scoped_logger_file_handler()

        return self._managed_resource

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType]
    ) -> Optional[bool]:
        self._managed_resource.context.free_scoped_logger_file_handler()
        # self.context.add_scoped_logger_file_handler()

        handled_err: bool = self._release_managed_subresource(self._managed_resource, exc_val)
        return handled_err

    @_with_method_as_context_manager
    def managed_subresource(self, *args, **kwds) -> SubAggregateT:
        self._managed_subresource(*args, **kwds)
        return self

    def _managed_subresource(self, *args, **kwds) -> ContextManager[SubAggregateT]:
        self._invariant()
        self._managed_resource: SubAggregateT = self._acquire_managed_subresource(
            *args, **kwds)
        return self

    @abstractmethod
    def _acquire_managed_subresource(self, *args, **kwds) -> SubAggregateT:
        """
        Acquire a new subaggregate that is managed such that it is deep saved at the
        beginning.
        """
        # Create subaggregate:
        with self.context.lock:
            self.refresh(self.is_deep)
            subdataclass: SubDataclassT = self.repo.load(*args, **kwds)
            subagg: SubAggregateT = self.subaggregate(subdataclass, self.context)
            self.save_subaggregate(subagg, deep=False)
            return subagg

    def _release_managed_subresource(self, resource: SubAggregateT, e: Exception = None) -> bool:
        """
        Release a subaggregate that was acquired with managed_subresource. The subaggregate
        normally has already saved itself. We may update things again here if needed.

        Exceptions may be handled here. If handled, return True, if not, then return False.
        """
        with self.context.lock:
            self.refresh(self.is_deep)
            self.save(False)
        handled_error = e is None
        return handled_error

    def __len__(self) -> int:
        return len(self._dataclass.get_sublocation())

    def __iter__(self) -> Iterable[SubAggregateT]:
        if not self.is_deep:
            self.refresh(True)
        for subdataclass in self._dataclass.get_sublocation_values():
            if subdataclass is not None:
                yield self.subaggregate(subdataclass, self.context, is_deep=True, parent=self)

    def __getitem__(self, item: int) -> 'Trial':
        """
        Get trial at the given index.

        :param item: trial index
        :return:
        """
        return self.subaggregate(self._dataclass.get_sublocation()[item], self.context, self.is_deep, self)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        if self.is_deep:
            prefix = "<deep>"
            subobjs = str([str(t) for t in self.__iter__()])
        else:
            prefix = "<shallow>"
            subobjs = f"{self._dataclass.get_sublocation()}>"
        return (
            f"{self.__class__.__name__}{prefix}("
            f"id={self._dataclass.get_id()}, {subobjs}"
            f")"
        )

    def __eq__(self, other: object) -> bool:
        return self._dataclass == other._dataclass


class Root(BaseAggregate[None, 'Project', RootDataclass]):

    def save(self, deep: bool = False):
        if deep:
            for p in self.projects:
                p.save(deep=p.is_deep)

    @property
    def projects(self) -> List['Project']:
        return list(self)

    @staticmethod
    def from_repo(context: CX, repo: HyperparamsRepository) -> 'Root':
        _dataclass: RootDataclass = repo.load(ScopedLocation())
        automl_context: AutoMLContext = AutoMLContext.from_context(context, repo)
        return Root(_dataclass, automl_context)

    @staticmethod
    def vanilla(context: CX = None) -> 'Root':
        if context is None:
            context = CX()
        vanilla_repo: HyperparamsRepository = VanillaHyperparamsRepository(
            os.path.join(context.get_path(), "hyperparams"))
        return Root.from_repo(context, vanilla_repo)

    @_with_method_as_context_manager
    def get_project(self, name: str) -> 'Root':
        self._managed_subresource(project_name=name)
        return self

    @_with_method_as_context_manager
    def default_project(self) -> 'Root':
        self._managed_subresource(project_name=DEFAULT_PROJECT)
        return self

    def _acquire_managed_subresource(self, project_name: str) -> 'Project':
        project_loc: ScopedLocation = self.loc.with_id(project_name)
        return super()._acquire_managed_subresource(project_loc)


class Project(BaseAggregate[None, 'Client', ProjectDataclass]):

    @_with_method_as_context_manager
    def get_client(self, name: str) -> 'Project':
        self._managed_subresource(client_name=name)
        return self

    @_with_method_as_context_manager
    def default_client(self) -> 'Project':
        self._managed_subresource(client_name=DEFAULT_CLIENT)
        return self

    def _acquire_managed_subresource(self, client_name: str) -> 'Client':
        client_loc: ScopedLocation = self.loc.with_id(client_name)
        return super()._acquire_managed_subresource(client_loc)


class Client(BaseAggregate[Project, 'Round', ClientDataclass]):

    @_with_method_as_context_manager
    def new_round(self, main_metric_name: str) -> 'Client':
        self._managed_subresource(new_round=True, main_metric_name=main_metric_name)
        return self

    @_with_method_as_context_manager
    def resume_last_round(self) -> 'Client':
        self._managed_subresource(new_round=False)
        return self

    @_with_method_as_context_manager
    def optim_round(self, new_round: bool, main_metric_name: str) -> 'Client':
        self._managed_subresource(new_round=new_round, main_metric_name=main_metric_name)
        return self

    def _acquire_managed_subresource(self, new_round: bool = True, main_metric_name: str = None) -> 'Round':
        with self.context.lock:
            self.refresh(self.is_deep)

            # Get new round loc:
            round_id: int = self._dataclass.get_next_i()
            if not new_round:
                round_id = max(0, round_id - 1)
            if round_id == 0:
                new_round = True
            round_loc: ScopedLocation = self.loc.with_id(round_id)

            # Get round to return:
            _round_dataclass: RoundDataclass = self.repo.load(round_loc, deep=True)
            if main_metric_name is not None:
                _round_dataclass.main_metric_name = main_metric_name

            subagg: Round = Round(_round_dataclass, self.context, is_deep=True)
            if new_round:
                self.save_subaggregate(subagg, deep=True)
            return subagg


class Round(BaseAggregate[Client, 'Trial', RoundDataclass]):

    def with_optimizer(
        self,
        hp_optimizer: BaseHyperparameterOptimizer,
        hp_space: HyperparameterSpace
    ) -> 'Round':
        self.hp_optimizer: BaseHyperparameterOptimizer = hp_optimizer
        self.hp_space: HyperparameterSpace = hp_space
        return self

    def with_metric(self, metric_name: str) -> 'Round':
        self._dataclass.main_metric_name = metric_name
        return self.save(False)

    @property
    def main_metric_name(self) -> str:
        return self._dataclass.main_metric_name

    @_with_method_as_context_manager
    def new_rvs_trial(self, continue_on_error: bool = False) -> 'Round':
        self._managed_subresource(new_trial=True, continue_on_error=continue_on_error)
        return self

    @_with_method_as_context_manager
    def last_trial(self, continue_on_error: bool = False) -> 'Round':
        self._managed_subresource(new_trial=False, continue_on_error=continue_on_error)
        return self

    @_with_method_as_context_manager
    def refitting_best_trial(self) -> 'Round':
        self.refresh(deep=True)
        self._managed_subresource(new_trial=None, continue_on_error=False)
        return self

    @property
    def _trials(self) -> List['Trial']:
        if not self.is_deep:
            self.refresh(True)
        return [Trial(t, self.context) for t in self._dataclass.get_sublocation()]

    @property
    def get_id(self) -> int:
        return self._dataclass.get_id()

    def _acquire_managed_subresource(self, new_trial=True, continue_on_error: bool = False) -> 'Trial':
        """
        Get a trial.

        :param new_trial: If True, will create a new trial. If false, will load the last trial. If None, will load the best trial.
        :param continue_on_error: If True, will continue to the next trial if the current trial fails.
                                  Otherwise, will let the exception be raised for the failure (won't catch).
        """
        with self.context.lock:
            self.refresh(self.is_deep)
            # self.context.add_scoped_logger_file_handler()

            # Get new trial loc:
            trial_id: int = self._dataclass.get_next_i()
            if new_trial is None:
                trial_id: int = self.get_best_trial()._dataclass.get_id()
            elif not new_trial:
                # Try get last trial
                trial_id = max(0, trial_id - 1)
            trial_loc = self.loc.with_id(trial_id)

            # Get trial to return:
            _trial_dataclass: TrialDataclass = self.repo.load(trial_loc, deep=True)
            new_hps: HyperparameterSamples = self.hp_optimizer.find_next_best_hyperparams(self)
            _trial_dataclass.hyperparams = new_hps
            self.flow.log_planned(new_hps)
            subagg: Trial = Trial(_trial_dataclass, self.context, is_deep=True)

            if continue_on_error:
                subagg.continue_loop_on_error()

            if new_trial or trial_id == 0:
                self.save_subaggregate(subagg, deep=True)
            return subagg

    def _release_managed_subresource(self, resource: 'Trial', e: Exception = None) -> bool:
        """
        Release a subaggregate that was acquired with managed_subresource. The subaggregate
        normally has already saved itself. We may update things again here if needed.

        Exceptions may be handled here. If handled, return True, if not, then return False.
        """
        resource: Trial = resource  # typing
        handled_exception = False
        with self.context.lock:
            self.refresh(True)

            is_all_success: bool = resource.are_all_splits_successful()
            is_all_failure: bool = resource.are_all_splits_failures()

            if e is None:
                handled_exception = True
                if is_all_success:
                    resource._set_success()
                elif is_all_failure:
                    e = RuntimeError("All trial splits failed for this trial.")
                    resource._set_failed(e)
            else:
                resource._set_failed(e)

            self.save_subaggregate(resource, deep=resource.is_deep)

            if not is_all_failure and len(self) > 0:
                main_metric_name = self.main_metric_name
                self.flow.log('Finished round hp search!')
                self.flow.log_best_hps(
                    main_metric_name,
                    self.get_best_hyperparams(main_metric_name)
                )
            else:
                self.flow.log_failure(e)

            self.save(False)
        return handled_exception

    def get_best_hyperparams(self, metric_name: str = None) -> HyperparameterSamples:
        """
        Get best hyperparams from all trials.

        :return:
        """
        if not self.is_deep:
            self.refresh(True)
        metric_name = self.sanitize_metric_name(metric_name)

        if len(self) == 0:
            return HyperparameterSamples()
        elif len(self) == 1:
            return self._trials[0].get_hyperparams()
        else:
            return self.get_best_trial(metric_name).get_hyperparams()

    def get_best_trial(self, metric_name: str = None) -> Optional['Trial']:
        """
        :return: trial with best score from all trials
        """
        best_score, best_trial = None, None
        metric_name = self.sanitize_metric_name(metric_name)

        if len(self) == 0:
            raise Exception('Could not get best trial because there were no successful trial.')

        for trial in self._trials:
            trial_score = trial.get_avg_validation_score(metric_name)

            _has_better_score = best_score is None or (
                trial_score is not None and trial.is_success() and (
                    self.is_higher_score_better(metric_name) == (trial_score > best_score)
                )
            )

            if _has_better_score:
                best_score = trial_score
                best_trial = trial

        return best_trial

    def is_higher_score_better(self, metric_name: str = None) -> bool:
        """
        Return true if higher score is better. If metric_name is None, the optimizer's
        metric is taken.

        :return
        """
        if len(self) == 0:
            return False

        metric_name = self.sanitize_metric_name(metric_name)

        is_higher_score_better_attr = "is_higher_score_better_attr_" + metric_name
        if not hasattr(self, is_higher_score_better_attr):
            setattr(
                self,
                is_higher_score_better_attr,
                self._trials[-1].is_higher_score_better(metric_name)
            )
        return getattr(self, is_higher_score_better_attr)

    def append(self, trial: 'Trial'):
        """
        Add a new trial. Will also save the trial shallowly.

        :param trial: new trial
        :return:
        """
        self.save_subaggregate(trial, deep=False)

    def copy(self) -> 'Round':
        return Round(copy.deepcopy(self._dataclass), self.context.copy().with_loc(self.loc.popped()))

    def get_number_of_split(self):
        if len(self) > 0:
            return len(self[0]._validation_splits)
        return 0

    def get_metric_names(self) -> List[str]:
        if len(self) > 0:
            return list(self[-1]._dataclass.validation_splits[-1].metric_results.keys())
        return []

    def best_result_summary(self, metric_name: str = None) -> Tuple[float, ScopedLocationAttrInt, FlatDict]:
        return self.summary(metric_name)[0]

    def summary(
        self, metric_name: str = None
    ) -> List[Tuple[float, ScopedLocationAttrInt, FlatDict]]:
        """
        Get a summary of the round. Best score is first.
        Values in the returned triplet tuples are: (score, trial_number, hyperparams),
        sorted by score such that the best score is first.
        """
        if not self.is_deep:
            self.refresh(True)
        metric_name = self.sanitize_metric_name(metric_name)

        results: List[float, ScopedLocationAttrInt, FlatDict] = list()

        for trial in self._trials:
            # TODO: what happens to failed or ongoing trials?
            score = trial.get_avg_validation_score(metric_name)
            trial_number = trial._dataclass.trial_number
            hp = trial.get_hyperparams().to_flat_dict()

            results.append((score, trial_number, hp))

        is_reverse: bool = self.is_higher_score_better(metric_name)
        results = list(sorted(results, reverse=is_reverse))
        return results

    def get_all_hyperparams(self) -> List[FlatDict]:
        """
        Get all hyperparams from all trials.

        :return:
        """
        self.refresh(True)

        hyperparams: List[FlatDict] = list()
        for trial in self._trials:
            hyperparams.append(trial.get_hyperparams().to_flat_dict())

        return hyperparams


class Trial(BaseAggregate[Round, 'TrialSplit', TrialDataclass]):
    """
    This class is a sub-contextualization of the :class:`HyperparameterRepository`
    class for holding a trial and manipulating it within its context.
    A Trial contains the results for each validation split.
    Each trial split contains both the training set results, and the validation set results.

    .. seealso::
        :class:`Trial`,
        :class:`TrialSplit`,
        :class:`HyperparamsRepository`,
        :class:`AutoML`,
        :class:`ExecutionContext`
    """

    def continue_loop_on_error(self) -> 'Trial':
        self._continue_loop_on_error: bool = True
        return self

    @property
    def _validation_splits(self) -> List['TrialSplit']:
        if not self.is_deep:
            self.refresh(True)
        return [TrialSplit(s, self.context, is_deep=self.is_deep)
                for s in self._dataclass.validation_splits]

    @property
    def main_metric_name(self) -> str:
        return self.parent.main_metric_name

    @_with_method_as_context_manager
    def new_validation_split(self, continue_loop_on_error: bool = False) -> 'Trial':
        continue_loop_on_error = continue_loop_on_error or (
            hasattr(self, "_continue_loop_on_error") and self._continue_loop_on_error)

        self._managed_subresource(continue_loop_on_error=continue_loop_on_error)
        return self

    def retrain_split(self) -> 'Trial':
        self._managed_subresource(continue_loop_on_error=False, retrain_split=True)
        return self

    def _acquire_managed_subresource(self, continue_loop_on_error: bool, retrain_split=False) -> 'TrialSplit':

        self.error_types_to_raise = (
            SystemError, SystemExit, EOFError, KeyboardInterrupt
        ) if continue_loop_on_error else (Exception,)

        with self.context.lock:
            self.refresh(self.is_deep)
            # self.context.add_scoped_logger_file_handler()

            # Get new split loc:
            if retrain_split:
                split_id = RETRAIN_TRIAL_SPLIT_ID
            else:
                split_id: int = self._dataclass.get_next_i()
                split_id = max(0, split_id)
                if split_id == 0:
                    self.flow.log_start()
            split_loc = self.loc.with_id(split_id)

            # Get split to return:
            _split_dataclass: TrialSplitDataclass = self.repo.load(split_loc)
            _split_dataclass.hyperparams = self.get_hyperparams()
            subagg: TrialSplit = TrialSplit(_split_dataclass, self.context, is_deep=True)

            self.save_subaggregate(subagg, deep=True)
            return subagg

    def _release_managed_subresource(self, resource: 'TrialSplit', e: Exception = None) -> bool:
        gc.collect()
        handled_error = False

        with self.context.lock:
            self.refresh(self.is_deep)
            # self.context.free_scoped_logger_handler_file()

            if e is None:
                resource._set_success()
                handled_error = True
            else:
                resource._set_failed(e)

                if any((isinstance(e, c) for c in self.error_types_to_raise)):
                    self._set_failed(e)
                    handled_error = False
                else:
                    handled_error = True

            if resource._dataclass.split_number == RETRAIN_TRIAL_SPLIT_ID:
                self._dataclass.retrained_split = resource._dataclass
                resource.save(deep=True)
                self.save(deep=False)
            else:
                self.save_subaggregate(resource, deep=True)

        return handled_error

    def is_success(self):
        """
        Checks if the trial is successful from its dataclass record.
        """
        return self._dataclass.status == TrialStatus.SUCCESS

    def get_status(self) -> TrialStatus:
        """
        Get the status of the trial.
        """
        return self._dataclass.status

    def are_all_splits_successful(self) -> bool:
        """
        Return true if all splits are successful.
        """
        return all(i.is_success() for i in self._validation_splits)

    def are_all_splits_failures(self) -> bool:
        """
        Return true if all splits are failed.
        """
        return all(not i.is_success() for i in self._validation_splits)

    def get_avg_validation_score(self, metric_name: str = None) -> Optional[float]:
        """
        Returns the average score for all validation splits's
        best validation score for the specified scoring metric.

        :return: validation score
        """
        if self.is_success():
            metric_name = self.sanitize_metric_name(metric_name)

            scores = [
                val_split[metric_name].get_best_validation_score()
                for val_split in self._validation_splits
                if val_split.is_success() and metric_name in val_split.get_metric_names()
            ]
            scores = [s for s in scores if s is not None]
            return sum(scores) / len(scores) if len(scores) > 0 else None

    def get_avg_n_epoch_to_best_validation_score(self, metric_name: str = None) -> float:
        metric_name = self.sanitize_metric_name(metric_name)

        n_epochs = [
            val_split.get_n_epochs_to_best_validation_score(metric_name)
            for val_split in self._validation_splits if val_split.is_success()
        ]

        n_epochs = sum(n_epochs) / len(n_epochs) if len(n_epochs) > 0 else None
        return n_epochs

    def get_hyperparams(self) -> HyperparameterSamples:
        """
        Return hyperparams dict.

        :return: hyperparams dict
        """
        return self._dataclass.hyperparams

    def _set_success(self) -> 'Trial':
        """
        Set trial status to success. Must save after to ensure consistency.

        :return: self
        """
        self._dataclass.end(TrialStatus.SUCCESS)
        self.flow.log_success()
        return self

    def _set_failed(self, error: Exception) -> 'Trial':
        """
        Set failed trial with exception. Must save after to ensure consistency.

        :param error: catched exception
        :return: self
        """
        self._dataclass.end(TrialStatus.FAILED)
        self.flow.log_failure(exception=error)
        return self

    def get_trial_id(self, hp_dict: Dict):
        """
        Hash hyperparams with blake2s to create a trial hash.

        :param hp_dict: hyperparams dict
        :return:
        """
        current_hyperparameters_hash = hashlib.blake2s(str.encode(str(hp_dict))).hexdigest()
        return f"{self._dataclass.trial_number}_{current_hyperparameters_hash}"

    def is_higher_score_better(self, metric_name: str = None) -> bool:
        """
        Return if the metric is higher score is better.

        :param metric_name: metric name
        :return: bool
        """
        metric_name = self.sanitize_metric_name(metric_name)
        return self._dataclass.validation_splits[-1].metric_results[metric_name].higher_score_is_better


class TrialSplit(BaseAggregate[Trial, 'MetricResults', TrialSplitDataclass]):
    """
    One split of a trial. This is where a model is trained and evaluated on a specific dataset split.
    """

    def __init__(self, _dataclass: SubDataclassT, context: AutoMLContext, is_deep=False, parent: Trial = None):
        if parent is not None:
            _dataclass.hyperparams = parent.get_hyperparams()
        super().__init__(_dataclass, context, is_deep=is_deep, parent=parent)
        self.epoch: int = 0
        self.n_epochs: int = None

    def _invariant(self):
        self._assert(self.is_deep,
                     f"self.is_deep should always be set to True for "
                     f"{self.__class__.__name__}", self.context)
        super()._invariant()

    @property
    def main_metric_name(self) -> str:
        return self.parent.main_metric_name

    def get_hyperparams(self) -> RecursiveDict:
        return self._dataclass.hyperparams

    def with_n_epochs(self, n_epochs: int) -> 'TrialSplit':
        self.n_epochs: int = n_epochs
        return self

    def next_epoch(self) -> int:
        """
        Increment epoch. Returns the new epoch id.
        Epochs are 1-indexed like lengths: first epoch is 1, second is 2, etc.
        """
        if self.n_epochs is None:
            raise ValueError(
                "self.n_epochs is not set. Please call self.with_n_epochs(n_epochs) first on your TrialSplit.")
        if self.epoch == 0:
            self.start()
        self.epoch: int = self.epoch + 1
        self.flow.log_epoch(self.epoch, self.n_epochs)
        return self.epoch

    def start(self) -> 'TrialSplit':
        """
        Start the trial split.
        """
        self.flow.log_start()
        with self.context.lock:
            self._dataclass.start()
            self.save(False)
        return self

    @property
    def metric_results(self) -> Dict[str, 'MetricResults']:
        return {
            metric_name: MetricResults(mr, self.context)  # .at_epoch(self.epoch, self.n_epochs)
            for metric_name, mr in self._dataclass.metric_results.items()
        }

    def metric_result(self, metric_name: str = None) -> 'MetricResults':
        """
        Get a metric result that is not managed with a "with" statement. Access it read-only.
        """
        metric_name = self.sanitize_metric_name(metric_name)
        mr: MetricResultsDataclass = self._dataclass.get_sublocation()[metric_name]
        return MetricResults(mr, self.context, is_deep=True)

    @_with_method_as_context_manager
    def managed_metric(self, metric_name: str, higher_score_is_better: bool) -> 'TrialSplit':
        """
        To be used as a with statement to get the managed metric.
        """
        self._managed_subresource(metric_name=metric_name, higher_score_is_better=higher_score_is_better)
        return self

    def _acquire_managed_subresource(self, metric_name: str, higher_score_is_better: bool) -> 'MetricResults':
        with self.context.lock:
            self.refresh(True)

            subdataclass: MetricResultsDataclass = self._create_or_get_metric_results(
                metric_name, higher_score_is_better)

            subagg: MetricResults = self.subaggregate(subdataclass, self.context, is_deep=True, parent=self)
            return subagg

    def _create_or_get_metric_results(self, name, higher_score_is_better):
        name = self.sanitize_metric_name(name)
        if name not in self._dataclass.metric_results:
            self._dataclass.metric_results[name] = MetricResultsDataclass(
                metric_name=name,
                validation_values=[],
                train_values=[],
                higher_score_is_better=higher_score_is_better,
            )
        return self._dataclass.metric_results[name]

    def _release_managed_subresource(self, resource: 'MetricResults', e: Exception = None) -> bool:
        handled_error = False
        with self.context.lock:
            if e is not None:
                handled_error = False
                self.context.flow.log_error(e)
            else:
                handled_error = True
            self.save_subaggregate(resource, deep=True)
        return handled_error

    def train_context(self) -> 'AutoMLContext':
        return self.context.train()

    def validation_context(self) -> 'AutoMLContext':
        return self.context.validation()

    def get_metric_names(self) -> List[str]:
        return list(self._dataclass.metric_results.keys())

    def _set_success(self) -> 'TrialSplit':
        """
        Set trial status to success. Must save after to ensure consistency.

        :return: self
        """
        self._dataclass.end(status=TrialStatus.SUCCESS)
        self.flow.log_success()
        return self

    def is_success(self):
        """
        Set trial status to success.
        """
        return self._dataclass.status == TrialStatus.SUCCESS

    def _set_failed(self, error: Exception) -> 'TrialSplit':
        """
        Set failed trial with exception. Must save after to ensure consistency.

        :param error: catched exception
        :return: self
        """
        if isinstance(error, SystemExit) or isinstance(error, KeyboardInterrupt):
            self._dataclass.end(TrialStatus.ABORTED)
            self.flow.log_aborted(error)
        else:
            self._dataclass.end(TrialStatus.FAILED)
            self.flow.log_failure(error)
        return self


class MetricResults(BaseAggregate[TrialSplit, None, MetricResultsDataclass]):

    def _invariant(self):
        self._assert(self.is_deep,
                     f"self.is_deep should always be set to True for "
                     f"{self.__class__.__name__}", self.context)
        super()._invariant()

    @property
    def metric_name(self) -> str:
        return self._dataclass.metric_name

    def _acquire_managed_subresource(self, *args, **kwds) -> SubAggregateT:
        raise NotImplementedError("MetricResults has no subresource to manage as a terminal resource.")

    def add_train_result(self, score: float):
        """
        Add a train metric result.

        :param name: name of the metric. If None, use the main metric name.
        :param score: the value to be logged
        :param higher_score_is_better: wheter or not a higher score is better for this metric
        """
        with self.context.lock:
            self._dataclass.train_values.append(score)
            self.flow.log_train_metric(self.metric_name, score)
            self.save(True)

    def add_valid_result(self, score: float):
        """
        Add a validation metric result.

        :param name: name of the metric. If None, use the main metric name.
        :param score: the value to be logged
        :param higher_score_is_better: wheter or not a higher score is better for this metric
        """
        with self.context.lock:
            self._dataclass.validation_values.append(score)
            self.flow.log_valid_metric(self.metric_name, score)
            self.save(True)

    def get_train_scores(self) -> List[float]:
        return self._dataclass.train_values

    def get_valid_scores(self) -> List[float]:
        """
        Return the validation scores for the given scoring metric.
        """
        return self._dataclass.validation_values

    def get_final_validation_score(self) -> float:
        """
        Return the latest validation score for the given scoring metric.
        """
        return self.get_valid_scores()[-1]

    def get_best_validation_score(self) -> float:
        """
        Return the best validation score for the given scoring metric.
        """
        scores = self.get_valid_scores()
        if len(scores) == 0:
            return None

        if self.is_higher_score_better():
            f = np.max
        else:
            f = np.min

        return f(scores)

    def get_n_epochs_to_best_validation_score(self) -> int:
        """
        Return the number of epochs
        """
        if self.is_higher_score_better():
            f = np.argmax
        else:
            f = np.argmin

        return f(self.get_valid_scores())

    def is_higher_score_better(self) -> bool:
        """
        Return True if higher scores are better for the main metric.

        :return:
        """
        return self._dataclass.higher_score_is_better

    def is_new_best_score(self) -> bool:
        """
        Return True if the latest validation score is the new best score.

        :return:
        """
        if self.get_best_validation_score() in self.get_valid_scores()[:-1]:
            return False
        return True

    def __iter__(self) -> Iterable[SubAggregateT]:
        """
        Loop over validation values.
        """
        return self._dataclass.validation_values


aggregate_2_subaggregate: typing.OrderedDict[Type[BaseAggregate], Type[BaseAggregate]] = OrderedDict([
    (type(None), Root),
    (Root, Project),
    (Project, Client),
    (Client, Round),
    (Round, Trial),
    (Trial, TrialSplit),
    (TrialSplit, MetricResults),
    (MetricResults, type(None)),
])

aggregate_2_dataclass: typing.OrderedDict[BaseAggregate, BaseDataclass] = OrderedDict([
    (Root, RootDataclass),
    (Client, ClientDataclass),
    (Round, RoundDataclass),
    (Project, ProjectDataclass),
    (Trial, TrialDataclass),
    (TrialSplit, TrialSplitDataclass),
    (MetricResults, MetricResultsDataclass),
])
dataclass_2_aggregate: typing.OrderedDict[BaseDataclass, BaseAggregate] = {
    dc: agg
    for agg, dc in aggregate_2_dataclass.items()
}



import numpy as np
from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import Trainer
from neuraxle.metaopt.callbacks import MetricCallback
from neuraxle.metaopt.data.aggregates import Round
from neuraxle.metaopt.validation import GridExplorationSampler, ValidationSplitter
from neuraxle.steps.misc import FitTransformCallbackStep, TapeCallbackFunction
from sklearn.metrics import mean_squared_error


def test_validation_splitter_handler_methods_should_split_data(tmpdir):
    transform_callback = TapeCallbackFunction()
    fit_callback = TapeCallbackFunction()
    pipeline = FitTransformCallbackStep(
        transform_callback_function=transform_callback,
        fit_callback_function=fit_callback,
        transform_function=lambda di: di * 2
    )
    metric: MetricCallback = MetricCallback("MSE", mean_squared_error, False)
    validation_split_wrapper = Trainer(
        callbacks=[metric],
        validation_splitter=ValidationSplitter(validation_size=0.1),
        n_epochs=1,
    )

    data_inputs = np.random.randint(low=1, high=100, size=(100, 5))
    expected_outputs = np.random.randint(low=1, high=100, size=(100, 5))
    dact = DACT(di=data_inputs, eo=expected_outputs)

    round_scope: Round = Round.dummy().with_metric(metric.name).save(deep=False)
    with round_scope.with_optimizer(GridExplorationSampler(), HyperparameterSpace()).new_rvs_trial() as trial_scope:
        trained_pipeline: FitTransformCallbackStep = validation_split_wrapper.train(
            pipeline, dact, trial_scope, return_trained_pipelines=True)[0]

    predicted_outputs = trained_pipeline.predict(data_inputs)
    fit_callback = trained_pipeline.fit_callback_function
    transform_callback = trained_pipeline.transform_callback_function

    assert np.array_equal(predicted_outputs, data_inputs * 2)

    # should fit on train split
    assert np.array_equal(fit_callback.data[0][0], data_inputs[0:90])
    assert np.array_equal(fit_callback.data[0][1], expected_outputs[0:90])

    # should transform on test split
    assert np.array_equal(transform_callback.data[0], data_inputs[0:90])
    assert np.array_equal(transform_callback.data[1], data_inputs[90:])

    # should predict on all data at the end
    assert np.array_equal(transform_callback.data[2], data_inputs)

    with round_scope.last_trial() as trial_scope:
        assert trial_scope.get_avg_validation_score() is not None



"""
Hyperparameter Dictionary Conversions
=====================================
Ways to convert from a nested dictionary of hyperparameters to a flat dictionary, and vice versa.

Here is a nested dictionary:

.. code-block:: python

    {
        "b": {
            "a": {
                "learning_rate": 7
            },
            "learning_rate": 9
        }
    }

Here is an equivalent flat dictionary for the previous nested one:

.. code-block:: python

    {
        "b.a.learning_rate": 7,
        "b.learning_rate": 9
    }

Notice that if you have a ``SKLearnWrapper`` on a sklearn Pipeline object, the hyperparameters past that point will use
double underscores ``__`` as a separator rather than a dot in flat dictionaries, and in nested dictionaries the
sklearn params will appear as a flat past the sklearn wrapper, which is fine.

By default, hyperparameters are stored inside a HyperparameterSpace or inside a HyperparameterSamples object, which
offers methods to do the conversions above, and also using ordered dicts (OrderedDict) to store parameters in-order.

A HyperparameterSpace can be sampled by calling the ``.rvs()`` method on it, which will recursively call ``.rvs()`` on all
the HyperparameterSpace and HyperparameterDistribution that it contains. It will return a HyperparameterSamples object.
A HyperparameterSpace can also be narrowed towards an better, finer subspace, which is itself a HyperparameterSpace.
This can be done by calling the ``.narrow_space_from_best_guess`` method of the HyperparameterSpace which will also
recursively apply the changes to its contained HyperparameterSpace and to all its contained HyperparameterDistribution.

The HyperparameterSamples contains sampled hyperparameter, that is, a valued point in the possible space. This is
ready to be sent to an instance of the pipeline to try and score it, for example.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, Iterable, Union
import typing

from neuraxle.hyperparams.distributions import (FixedHyperparameter,
                                                HPSampledValue,
                                                HyperparameterDistribution)
from neuraxle.hyperparams.scipy_distributions import (
    ScipyContinuousDistributionWrapper, ScipyDiscreteDistributionWrapper)
from scipy.stats import rv_continuous, rv_discrete
from scipy.stats._distn_infrastructure import rv_generic


FlatDict = typing.OrderedDict[str, HPSampledValue]
RecursiveDictValue = Union[Any, 'RecursiveDict']


# class RecursiveDict(OrderedDict[str, RecursiveDictValue]):
class RecursiveDict(OrderedDict):
    """
    A data structure that provides an interface to access nested dictionaries with "flattened keys", and a few more functions.

    e.g.
        dct = RecursiveDict({'a':{'b':2}})
        assert dct["a__b"] == 2
        dct["a__b__c"] = 3
        assert dct['a']['b']['c'] == dct["a__b__c"] == dct.to_flat_dict()["a__b__c"]

    This class serves as a base for HyperparameterSamples and HyperparameterSpace
    """
    DEFAULT_SEPARATOR: str = '__'

    def __init__(self, *args, separator=None, **kwds):

        if separator is None:
            if self._is_only_arg_a_recursive_dict(args, kwds):
                separator = args[0].separator
            else:
                separator = self.DEFAULT_SEPARATOR
        self.separator = separator

        OrderedDict.__init__(self)
        for arg in args:
            self.update(arg)
        self.update(kwds)
        self._patch_args()

    def _is_only_arg_a_recursive_dict(self, args, kwds):
        return len(args) == 1 and isinstance(args[0], RecursiveDict) and len(kwds) == 0

    def _patch_args(self):
        to_patch_key_values = []
        for k, v in self.items():
            if isinstance(v, RecursiveDict):
                v._patch_args()
            else:
                patched_arg, did_patch = self._patch_arg(v)
                if did_patch:
                    to_patch_key_values.append((k, patched_arg))

        for k, v in to_patch_key_values:
            self[k] = v

    def same_class_new_instance(self, *args, **kwds):
        return type(self)(*args, separator=self.separator, **kwds)

    def _patch_arg(self, arg):
        """
        Patches argument if needed.
        :param arg: arg to patch if needed.
        :return: (patched_arg, did_patch)
        """
        return arg, False

    def get(self, key) -> RecursiveDictValue:
        try:
            return self[key]
        except KeyError:
            return self.same_class_new_instance()

    def __getitem__(self, key) -> RecursiveDictValue:
        return self._rec_get(key)

    def _rec_get(self, key) -> RecursiveDictValue:
        """
        Split the keys and call getter recursively until we get to the desired element.
        None returns every non-recursive elements.
        """
        if key is None:
            return self.get_root_leaf_data()

        lkey, _, rkey = key.partition(self.separator)
        rec_dict: RecursiveDict = OrderedDict.__getitem__(self, lkey)
        if rkey == "":
            return rec_dict
        else:
            # Splitted on sep and recursively call getter
            return rec_dict._rec_get(rkey)

    def get_root_leaf_data(self) -> FlatDict:
        """
        Returns a dictionary of all the non-recursive elements.
        That is, all the elements that are not RecursiveDict in the
        current root OrderedDict.
        """
        return dict(filter(lambda x: not isinstance(x[1], RecursiveDict), self.items()))

    def __setitem__(self, key: str, value: RecursiveDictValue):
        lkey, _, rkey = key.partition(self.separator)
        if rkey == "":
            if isinstance(value, dict) and not isinstance(value, RecursiveDict):
                value = self.same_class_new_instance(value.items())
            OrderedDict.__setitem__(self, lkey, value)
        else:
            if lkey not in self:
                OrderedDict.__setitem__(self, lkey, self.same_class_new_instance())
            self[lkey][rkey] = value

    def __contains__(self, key: str) -> bool:
        try:
            _ = self[key]
            return True
        except KeyError:
            return False

    def iter_flat(self, pre_key="", values_only=False) -> Iterable:
        """
        Returns a generator which yield (flatenned_key, value) pairs. value is never a RecursiveDict instance.
        Keys are sorted, then values are sorted as well.
        """
        for k, v in sorted(self.items()):
            if isinstance(v, RecursiveDict):
                yield from v.iter_flat(pre_key + k + self.separator, values_only=values_only)
            else:
                if values_only:
                    yield v
                else:
                    yield (pre_key + k, v)

    def to_flat_dict(self) -> FlatDict:
        """
        Returns a FlatDict, that is a totally flatened OrderedDict[str, HPSampledValue],
        with no recursively nested elements, i.e.: {fully__flattened__params: value}.

        .. info::
            The returned FlatDict is sorted in a new alphabetical order.

        """
        return OrderedDict(self.iter_flat())

    def to_nested_dict(self) -> Dict[str, RecursiveDictValue]:
        """
        Returns a dictionary counterpart which is still nested but that contains no RecursiveDict.
        """
        out_dict = dict()
        for k, v in self.items():
            if isinstance(v, RecursiveDict):
                v = v.to_nested_dict()
            out_dict[k] = v
        return out_dict

    def with_separator(self, separator: str):
        """
        Create a new recursive dict (from copy) that uses the given separator at each level.
        """
        return type(self)(
            separator=separator, **{
                key: value
                if not isinstance(value, RecursiveDict)
                else value.with_separator(separator) for key, value in self.items()
            })

    def is_empty(self) -> bool:
        return len(self) == 0 or len(self.to_flat_dict()) == 0


class HyperparameterSamples(RecursiveDict):
    """
    Wraps an hyperparameter nested dict or flat dict, and offer a few more functions.

    This can be set on a Pipeline with the method ``set_hyperparams``.

    HyperparameterSamples are often the result of calling ``.rvs()`` on an HyperparameterSpace.
    """

    def __init__(self, *args, separator=None, **kwds):
        super().__init__(*args, separator=separator, **kwds)


class HyperparameterSpace(RecursiveDict):
    """
    Wraps an hyperparameter nested dict or flat dict, and offer a few more functions to process
    all contained HyperparameterDistribution.

    This can be set on a Pipeline with the method ``set_hyperparams_space``.

    Calling ``.rvs()`` on an ``HyperparameterSpace`` results in ``HyperparameterSamples``.
    """

    def __init__(self, *args, separator=None, **kwds):
        super().__init__(*args, separator=separator, **kwds)

    def _patch_arg(self, arg):
        """
        Override of the RecursiveDict's default _patch_arg(arg) method.

        :param arg: arg to patch if needed.
        :return: (patched_arg, did_patch)
        """
        did_patch = False
        if hasattr(arg, 'dist') and isinstance(arg.dist, rv_generic):
            if isinstance(arg.dist, rv_continuous):
                arg, did_patch = ScipyContinuousDistributionWrapper(arg), True
            elif isinstance(arg.dist, rv_discrete):
                arg, did_patch = ScipyDiscreteDistributionWrapper(arg), True
        assert isinstance(arg, HyperparameterDistribution), (
            f"Hyperparameter space's distributions must be a dict of `HyperparameterDistributions`. "
            f"got `{arg}` of type `{type(arg)}` instead. You might consider using a "
            f"{FixedHyperparameter.__name__} class like {FixedHyperparameter}({arg}) to have a fixed value."
        )
        return arg, did_patch

    def rvs(self) -> 'HyperparameterSamples':
        """
        Sample the space of random variables.

        :return: a random HyperparameterSamples, sampled from a point of the present HyperparameterSpace.
        """
        new_items = []
        for k, v in self.iter_flat():
            if isinstance(v, HyperparameterDistribution):
                v = v.rvs()
            new_items.append((k, v))
        return HyperparameterSamples(new_items)

    # TODO : The following functions aren't really used, or tested. They should work though.

    def nullify(self):
        new_items = []
        for k, v in self.iter_flat():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                v = v.nullify()
            new_items.append((k, v))
        return HyperparameterSamples(new_items)

    def narrow_space_from_best_guess(
            self,
            best_guesses: 'HyperparameterSpace',
            kept_space_ratio: float = 0.5
    ) -> 'HyperparameterSpace':
        """
        Takes samples estimated to be the best ones of the space as of yet, and restrict the whole space towards that.

        :param best_guesses: sampled HyperparameterSpace (the result of rvs on each parameter, but still stored as a HyperparameterSpace).
        :param kept_space_ratio: what proportion of the space is kept. Should be between 0.0 and 1.0. Default is 0.5.
        :return: a new HyperparameterSpace containing the narrowed HyperparameterDistribution objects.
        """
        new_items = []
        for k, v in self.iter_flat():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                best_guess_v = best_guesses[k]
                v = v.narrow_space_from_best_guess(best_guess_v, kept_space_ratio)
            new_items.append((k, v))
        return HyperparameterSpace(new_items)

    def unnarrow(self) -> 'HyperparameterSpace':
        """
        Return the original space before narrowing of the distribution. If the distribution was never narrowed,
        the values in the dict will be copies.

        :return: the original HyperparameterSpace before narrowing.
        """
        new_items = []
        for k, v in self.iter_flat():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                v = v.unnarrow()
            new_items.append((k, v))
        return HyperparameterSpace(new_items)



"""
Neuraxle Test Case Class
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import numpy as np

from neuraxle.base import ExecutionMode


class NeuraxleTestCase:
    def __init__(
            self,
            pipeline,
            callbacks,
            expected_callbacks_data,
            hyperparams_space=None,
            hyperparams=None,
            expected_processed_outputs=None,
            execution_mode=None,
            more_arguments=None,
            data_inputs=None,
            expected_outputs=None
    ):
        self.expected_outputs = expected_outputs
        self.data_inputs = data_inputs
        self.execution_mode = execution_mode
        self.pipeline = pipeline
        self.callbacks = callbacks
        self.expected_callbacks_data = expected_callbacks_data
        self.hyperparams = hyperparams
        self.hyperparams_space = hyperparams_space
        self.expected_processed_outputs = expected_processed_outputs
        self.more_arguments = more_arguments

    def assert_callback_data_is_as_expected(self):
        for callback, expected_callback_data in zip(self.callbacks, self.expected_callbacks_data):
            if len(callback.data) > 0:
                if isinstance(callback.data[0], tuple):
                    for (expected_di, expected_eo), (actual_di, actual_eo) in zip(expected_callback_data, callback.data):
                        assert np.array_equal(expected_di, actual_di)
                        assert np.array_equal(expected_eo, actual_eo)
                else:
                    assert np.array_equal(
                        np.array(callback.data),
                        expected_callback_data
                    )
            else:
                assert np.array_equal(
                    np.array([]),
                    np.array(expected_callback_data)
                )

    def assert_expected_processed_outputs(self, processed_outputs):
        if self.execution_mode != ExecutionMode.FIT:
            assert np.array_equal(processed_outputs, self.expected_processed_outputs)

    def execute(self):
        for c in self.callbacks:
            c.reset()

        processed_outputs = None
        if self.execution_mode == ExecutionMode.TRANSFORM:
            processed_outputs = self.pipeline.transform(self.data_inputs)
        if self.execution_mode == ExecutionMode.FIT_TRANSFORM:
            self.pipeline, processed_outputs = self.pipeline.fit_transform(self.data_inputs, self.expected_outputs)
        if self.execution_mode == ExecutionMode.FIT:
            self.pipeline = self.pipeline.fit(self.data_inputs, self.expected_outputs)

        return processed_outputs



"""
Neuraxle's SQLAlchemy Hyperparameter Repository Classes
=================================================
Data objects and related repositories used by AutoML, SQL version.

Classes are splitted like this for the AutoML:
- Projects
- Clients
- Rounds (runs)
- Trials
- TrialSplits
- MetricResults

..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import json
import os
from copy import deepcopy
from typing import List

from neuraxle.logging.logging import NeuraxleLogger
from neuraxle.metaopt.data.vanilla import (BaseDataclass,
                                           HyperparamsRepository,
                                           ScopedLocation, SubDataclassT,
                                           VanillaHyperparamsRepository,
                                           dataclass_2_id_attr, from_json,
                                           to_json)

ON_DISK_DELIM: str = "_"


class _OnDiskRepositoryLoggerHandlerMixin:
    """
    Mixin to add a disk logging handler to a repository. It has a cache_folder.
    """

    def __init__(self, cache_folder: str):
        self.cache_folder = cache_folder

    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        """
        Adds an on-disk logging handler to the repository.
        The file at this scope can be retrieved with the method :func:`get_scoped_logger_path`.
        """
        logging_file = self.get_scoped_logger_path(scope)
        os.makedirs(os.path.dirname(logging_file), exist_ok=True)
        logger.with_file_handler(logging_file)
        return self

    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        return ''.join(logger.read_log_file())

    def get_folder_at_scope(self, scope: ScopedLocation) -> str:
        _scope_attrs = scope.as_list(stringify=True)
        _scope_attrs = [ON_DISK_DELIM + s for s in _scope_attrs]
        return os.path.join(self.cache_folder, *_scope_attrs)

    def get_scoped_logger_path(self, scope: ScopedLocation) -> str:
        scoped_path: str = self.get_folder_at_scope(scope)
        return os.path.join(scoped_path, 'log.txt')


class HyperparamsOnDiskRepository(_OnDiskRepositoryLoggerHandlerMixin, HyperparamsRepository):
    """
    Hyperparams repository that saves json files for every AutoML trial.

    .. seealso::
        :class:`AutoML`,
        :class:`Trainer`,
    """

    def __init__(self, cache_folder: str = None):
        HyperparamsRepository.__init__(self)
        _OnDiskRepositoryLoggerHandlerMixin.__init__(self, cache_folder=cache_folder)
        self._vanilla = VanillaHyperparamsRepository(cache_folder=cache_folder)

    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        """
        Get metadata from scope.

        The fetched metadata will be the one that is the last item
        that is not a None in the provided scope.

        :param scope: scope to get metadata from.
        :return: metadata from scope.
        """
        loaded = self._load_dc(scope=scope, deep=deep)
        self._vanilla.save(loaded, scope=scope, deep=deep)
        return loaded

    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'HyperparamsRepository':
        """
        Save metadata to scope.

        :param metadata: metadata to save.
        :param scope: scope to save metadata to.
        :param deep: if True, save metadata's sublocations recursively so as to update.
        """
        self._vanilla.save(_dataclass=_dataclass, scope=scope, deep=deep)
        self._save_dc(_dataclass=_dataclass, scope=scope, deep=deep)
        return self

    def _load_dc(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        scope, _, load_file = self._get_dataclass_filename_path(None, scope)

        if not os.path.exists(load_file):
            # raise FileNotFoundError(f"{load_file} not found.")
            return self._vanilla.load(scope=scope, deep=deep)

        with open(load_file, 'r') as f:
            _dataclass: SubDataclassT = from_json(json.load(f))
            if _dataclass.has_sublocation_dataclasses():
                _dataclass = self._load_dc_sublocation_keys(_dataclass, scope)
                if deep is True:
                    for sub_dc_id in _dataclass.get_sublocation_keys():
                        sub_dc = self._load_dc(scope=scope.with_id(sub_dc_id), deep=deep)
                        _dataclass.store(sub_dc)
            return _dataclass

    def _load_dc_sublocation_keys(self, _dataclass: SubDataclassT, scope) -> SubDataclassT:
        dc_folder: str = self.get_folder_at_scope(scope)
        sublocs: List[str] = os.listdir(dc_folder)
        sublocs = [s[len(ON_DISK_DELIM):] for s in sublocs if s.startswith(ON_DISK_DELIM)]
        _dataclass.set_sublocation_keys(sublocs)
        return _dataclass

    def _save_dc(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False):
        scope, save_folder, save_file = self._get_dataclass_filename_path(_dataclass, scope)

        os.makedirs(save_folder, exist_ok=True)
        with open(save_file, 'w') as f:
            json.dump(to_json(_dataclass.empty()), f, indent=4)

        if deep is True and _dataclass.has_sublocation_dataclasses():
            for sub_dc in _dataclass.get_sublocation_values():
                self._save_dc(sub_dc, scope=scope, deep=deep)

    def _get_dataclass_filename_path(self, _dataclass: SubDataclassT, scope: ScopedLocation):
        scope = self._patch_scope_for_dataclass(_dataclass, scope)

        save_folder = self.get_folder_at_scope(scope)
        save_file = os.path.join(save_folder, 'metadata.json')

        return scope, save_folder, save_file

    def _patch_scope_for_dataclass(self, _dataclass: BaseDataclass, scope: ScopedLocation):
        scope = deepcopy(scope)
        if _dataclass is not None and _dataclass.get_id() is not None:
            scope = scope.at_dc(_dataclass)
            setattr(scope, dataclass_2_id_attr[_dataclass.__class__], _dataclass.get_id())
        return scope



import numpy as np

from neuraxle.data_container import ZipDataContainer, DACT


def test_zip_data_container_should_merge_two_data_sources_together():
    data_inputs_3d, expected_outputs_3d = _create_data_source((10, 10, 2))
    data_inputs_2d, expected_outputs_2d = _create_data_source((10, 10))
    data_container_2d = DACT(data_inputs=data_inputs_2d, expected_outputs=expected_outputs_2d)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d)

    zip_data_container = ZipDataContainer.create_from(data_container, data_container_2d)

    assert zip_data_container.ids == data_container.ids
    for i, di in enumerate(zip_data_container.data_inputs):
        assert np.array_equal(di[0], data_inputs_3d[i])
        assert np.array_equal(di[1], data_inputs_2d[i])


def test_zip_data_container_should_merge_1d_with_2d():
    data_inputs_3d, expected_outputs_3d = _create_data_source((10, 10, 2))
    data_inputs_1d, expected_outputs_1d = _create_data_source((10,))
    data_container_1d = DACT(data_inputs=data_inputs_1d, expected_outputs=expected_outputs_1d)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d)

    zip_data_container = ZipDataContainer.create_from(data_container, data_container_1d)

    assert zip_data_container.ids == data_container.ids
    for i, di in enumerate(zip_data_container.data_inputs):
        assert np.array_equal(di[0], data_inputs_3d[i])
        assert np.array_equal(di[1], data_inputs_1d[i])


def test_zip_data_container_should_merge_multiple_data_sources_together():
    data_inputs_3d, expected_outputs_3d = _create_data_source((10, 10, 2))
    data_inputs_2d, expected_outputs_2d = _create_data_source((10, 10))
    data_inputs_1d, expected_outputs_1d = _create_data_source((10,))
    data_container_1d = DACT(data_inputs=data_inputs_1d, expected_outputs=expected_outputs_1d)
    data_container_2d = DACT(data_inputs=data_inputs_2d, expected_outputs=expected_outputs_2d)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d)

    zip_data_container = ZipDataContainer.create_from(data_container, data_container_2d, data_container_1d)

    assert zip_data_container.ids == data_container.ids
    for i, di in enumerate(zip_data_container.data_inputs):
        assert np.array_equal(di[0], data_inputs_3d[i])
        assert np.array_equal(di[1], data_inputs_2d[i])


def test_zip_data_container_should_concatenate_inner_features():
    data_inputs_3d, expected_outputs_3d = _create_data_source((10, 10, 2))
    data_inputs_2d, expected_outputs_2d = _create_data_source((10, 10))
    data_container_2d = DACT(data_inputs=data_inputs_2d, expected_outputs=expected_outputs_2d)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d)

    zip_data_container = ZipDataContainer.create_from(data_container, data_container_2d)
    zip_data_container.concatenate_inner_features()

    assert np.array_equal(np.array(zip_data_container.data_inputs)[..., -1], data_container_2d.data_inputs)
    assert np.array_equal(np.array(zip_data_container.expected_outputs), expected_outputs_3d)


def _create_data_source(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs



import numpy as np
import pytest

from neuraxle.base import ExecutionMode
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import ReversiblePreprocessingWrapper
from neuraxle.steps.misc import TapeCallbackFunction, CallbackWrapper
from neuraxle.steps.numpy import MultiplyByN, AddN
from testing.steps.neuraxle_test_case import NeuraxleTestCase

DATA_INPUTS = np.array(range(5))
EXPECTED_OUTPUTS = np.array(range(5, 10))
EXPECTED_PROCESSED_OUTPUTS = np.array([5.0, 6.0, 7.0, 8.0, 9.0])

tape_transform_preprocessing = TapeCallbackFunction()
tape_fit_preprocessing = TapeCallbackFunction()
tape_transform_postprocessing = TapeCallbackFunction()
tape_fit_postprocessing = TapeCallbackFunction()
tape_inverse_transform_preprocessing = TapeCallbackFunction()


@pytest.mark.parametrize('test_case', [
    NeuraxleTestCase(
        pipeline=Pipeline([
            ReversiblePreprocessingWrapper(
                preprocessing_step=CallbackWrapper(MultiplyByN(2), tape_transform_preprocessing, tape_fit_postprocessing, tape_inverse_transform_preprocessing),
                postprocessing_step=CallbackWrapper(AddN(10), tape_transform_postprocessing, tape_fit_postprocessing)
            )]
        ),
        callbacks=[tape_transform_preprocessing, tape_fit_preprocessing, tape_transform_postprocessing, tape_fit_postprocessing, tape_inverse_transform_preprocessing],
        expected_callbacks_data=[
            [DATA_INPUTS],
            [],
            [DATA_INPUTS * 2],
            [],
            [(DATA_INPUTS * 2) + 10]
        ],
        data_inputs=DATA_INPUTS,
        expected_processed_outputs=EXPECTED_PROCESSED_OUTPUTS,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=Pipeline([
            ReversiblePreprocessingWrapper(
                preprocessing_step=CallbackWrapper(MultiplyByN(2), tape_transform_preprocessing, tape_fit_preprocessing, tape_inverse_transform_preprocessing),
                postprocessing_step=CallbackWrapper(AddN(10), tape_transform_postprocessing, tape_fit_postprocessing)
            )]
        ),
        callbacks=[tape_transform_preprocessing, tape_fit_preprocessing, tape_transform_postprocessing, tape_fit_postprocessing, tape_inverse_transform_preprocessing],
        expected_callbacks_data=[
            [DATA_INPUTS],
            [(DATA_INPUTS, EXPECTED_OUTPUTS)],
            [DATA_INPUTS * 2],
            [(DATA_INPUTS * 2, EXPECTED_OUTPUTS)],
            [(DATA_INPUTS * 2) + 10]
        ],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_processed_outputs=EXPECTED_PROCESSED_OUTPUTS,
        execution_mode=ExecutionMode.FIT_TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=Pipeline([
            ReversiblePreprocessingWrapper(
                preprocessing_step=CallbackWrapper(MultiplyByN(2), tape_transform_preprocessing, tape_fit_preprocessing, tape_inverse_transform_preprocessing),
                postprocessing_step=CallbackWrapper(AddN(10), tape_transform_postprocessing, tape_fit_postprocessing)
            )]
        ),
        callbacks=[tape_transform_preprocessing, tape_fit_preprocessing, tape_transform_postprocessing, tape_fit_postprocessing, tape_inverse_transform_preprocessing],
        expected_callbacks_data=[
            [DATA_INPUTS],
            [(DATA_INPUTS, EXPECTED_OUTPUTS)],
            [],
            [(DATA_INPUTS * 2, EXPECTED_OUTPUTS)],
            []
        ],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        execution_mode=ExecutionMode.FIT
    )
])
def test_reversible_preprocessing_wrapper(test_case):
    processed_outputs = test_case.execute()

    test_case.assert_expected_processed_outputs(processed_outputs)
    test_case.assert_callback_data_is_as_expected()



import argparse

import urllib3

# Here is the command I use to verify all libraries installed with pip.
# python license_checker_v2.py --dependencies $(cut -d '=' -f 1 <<< $(pip freeze))
# Alternatively, if you are processing directly a requirements.txt file, you can use the following command to parse
# awk -F'[>=<]' '{print $1}' requirements.txt

parser = argparse.ArgumentParser()
parser.add_argument('--dependencies', nargs='+', required=True,
                    help="A list of python library name you want to check the license of.")
parser.add_argument('--accepted_licenses', nargs='*',
                    help="A list of license which are considered acceptable for your project.",
                    default=["Apache Software License", "Apache 2.0", "BSD", "ZLIB", "MIT", "Unlicense", "CC0", "CC-BY","PSF", "MPL", "Mozilla Public License 2.0", "Historical Permission Notice and Disclaimer", "HPND"])
parser.add_argument('--forbidden_licenses', nargs='*',
                    help="A list of license which are considered problematic for your project.",
                    default=["GNU", "GPL", "Commons Clause", "BY-N"])
args = parser.parse_args()

python_dependencies = args.dependencies
pypi_pages = {}

for library_name in python_dependencies:
    url = f"https://pypi.org/project/{library_name}/"
    http_pool = urllib3.connection_from_url(url)
    result = http_pool.urlopen('GET', url)
    html_page = result.data.decode('utf-8')
    pypi_pages[library_name] = html_page


def parse_html(html_page):
    lines = html_page.split('\n')

    for i, l in enumerate(lines):
        if ("<a href=" in l) and ("License:" in l):
            return lines[i + 1].split('::')[1].strip('\n')

    for i, l in enumerate(lines):
        if "<strong>License:</strong> " in l:
            return lines[i].replace("<p><strong>License:</strong> ", "").replace('</p>', '')

    raise ValueError("Unable to find license in html page")


unknown_licenses = []
library_license_dict = {}
accepted_libraries = []
refused_libraries = []
maybe_libraries = []


def is_license_in_list(license, license_list):
    for l in license_list:
        if l.lower() in license.lower():
            return True
    return False


for library_name in python_dependencies:
    try:
        library_license = parse_html(pypi_pages[library_name])
        library_license_dict[library_name] = library_license
        print(f"{library_name}: {library_license}")
        # First checks if its refused_licenses, then if its in accepted_licenses, else add in the maybe list

        if is_license_in_list(library_license, args.forbidden_licenses):
            refused_libraries.append(library_name)
        elif is_license_in_list(library_license, args.accepted_licenses):
            accepted_libraries.append(library_name)
        else:
            maybe_libraries.append(library_name)

    except Exception as e:
        print(f"{library_name}: {e}")
        unknown_licenses.append(library_name)


def plurial(lst, _if='s', _else=''):
    return _if if len(lst) > 1 else _else

if len(unknown_licenses) > 0:
    print(f"Couldn't find the license{plurial(unknown_licenses)} of the following dependencies: {unknown_licenses}")

print(f"\nThe following dependenc{plurial(accepted_libraries, 'y', 'ies')} have an accepted license: {accepted_libraries}")

if len(refused_libraries) > 0:
    print(f"The following dependencie{plurial(refused_libraries, 'y', 'ies')} have forbidden license(s):")
    for library_name in refused_libraries:
        print(f"  {library_name}: {library_license_dict[library_name]}")

if len(maybe_libraries) > 0:
    print(f"The following dependencie{plurial(maybe_libraries, 'y', 'ies')} have license which needs to be reviewed: ")
    for library_name in maybe_libraries:
        print(f"  {library_name}: {library_license_dict[library_name]}")


assert len(refused_libraries) == 0 and len(maybe_libraries) == 0 and len(unknown_licenses) == 0




import copy
from collections import defaultdict, namedtuple
from dataclasses import dataclass, field
from typing import Dict, List, Set, Tuple, Type, Any

import numpy as np
import pytest
from neuraxle.base import (CX, BaseStep, HandleOnlyMixin, Identity, MetaStep,
                           NonFittableMixin)
from neuraxle.data_container import DataContainer as DACT
from neuraxle.data_container import PredsDACT, TrainDACT
from neuraxle.distributed.streaming import ParallelQueuedFeatureUnion
from neuraxle.hyperparams.distributions import (
    DiscreteHyperparameterDistribution, PriorityChoice, RandInt, Uniform)
from neuraxle.hyperparams.space import (FlatDict, HyperparameterSamples,
                                        HyperparameterSpace)
from neuraxle.metaopt.auto_ml import ControlledAutoML, DefaultLoop, Trainer
from neuraxle.metaopt.callbacks import MetricCallback
from neuraxle.metaopt.data.aggregates import Round
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, AutoMLContext,
                                           BaseDataclass, RoundDataclass,
                                           ScopedLocation)
from neuraxle.metaopt.validation import (GridExplorationSampler,
                                         ValidationSplitter)
from neuraxle.pipeline import Pipeline
from neuraxle.steps.data import DataShuffler
from neuraxle.steps.flow import TrainOnlyWrapper
from neuraxle.steps.misc import AssertFalseStep, Sleep
from neuraxle.steps.numpy import AddN
from sklearn.metrics import median_absolute_error


class StepThatAssertsContextIsSpecifiedAtTrain(Identity):
    def __init__(self, expected_loc: ScopedLocation, up_to_dc: Type[BaseDataclass] = RoundDataclass):
        BaseStep.__init__(self)
        HandleOnlyMixin.__init__(self)
        self.expected_loc = expected_loc
        self.up_to_dc: Type[BaseDataclass] = up_to_dc

    def _did_process(self, data_container: DACT, context: CX) -> DACT:
        if self.is_train:
            context: AutoMLContext = context  # typing annotation for IDE
            self._assert_equals(
                self.expected_loc[:self.up_to_dc], context.loc[:self.up_to_dc],
                f'Context is not at the expected location. '
                f'Expected {self.expected_loc}, got {context.loc}.',
                context)
            self._assert_equals(
                context.loc in context.repo.root, True,
                "Context should have the dataclass, but it doesn't", context)
        return data_container


def test_automl_context_is_correctly_specified_into_trial_with_full_automl_scenario(tmpdir):
    # This is a large test
    dact = DACT(di=list(range(10)), eo=list(range(10, 20)))
    cx = AutoMLContext.from_context(CX(root=tmpdir))
    expected_deep_cx_loc = ScopedLocation.default(0, 0, 0)
    assertion_step = StepThatAssertsContextIsSpecifiedAtTrain(expected_loc=expected_deep_cx_loc)
    automl: ControlledAutoML[Pipeline] = _create_automl_test_loop(tmpdir, assertion_step)
    automl = automl.handle_fit(dact, cx)

    pred: DACT = automl.handle_predict(dact.without_eo(), cx)

    round_scope: Round = cx.with_loc(ScopedLocation.default(round_number=0)).load_agg()
    best: Tuple[float, int, FlatDict] = round_scope.best_result_summary()
    best_score = best[0]
    assert best_score == 0
    best_add_n: int = list(best[-1].values())[0]
    assert best_add_n == 10  # We expect the AddN step to make use of the value "10" for 0 MAE error.
    assert np.array_equal(list(pred.di), list(dact.eo))


def test_automl_step_can_interrupt_on_fail_with_full_automl_scenario(tmpdir):
    # This is a large test
    dact = DACT(di=list(range(10)), eo=list(range(10, 20)))
    cx = CX(root=tmpdir)
    assertion_step = AssertFalseStep()
    automl = _create_automl_test_loop(tmpdir, assertion_step)

    with pytest.raises(AssertionError):
        automl.handle_fit(dact, cx)


def _create_automl_test_loop(tmpdir, assertion_step: BaseStep, n_trials: int = 4, start_new_round=True, refit_best_trial=True):
    automl = ControlledAutoML(
        pipeline=Pipeline([
            TrainOnlyWrapper(DataShuffler()),
            AddN().with_hp_range(range(8, 12)),
            assertion_step
        ]),
        loop=DefaultLoop(
            trainer=Trainer(
                validation_splitter=ValidationSplitter(validation_size=0.2),
                n_epochs=4,
                callbacks=[MetricCallback('MAE', median_absolute_error, False)],
            ),
            hp_optimizer=GridExplorationSampler(n_trials),
            n_trials=n_trials,
            continue_loop_on_error=False,
        ),
        main_metric_name='MAE',
        start_new_round=start_new_round,
        refit_best_trial=refit_best_trial,
    )

    return automl


@pytest.mark.parametrize('n_trials', [1, 3, 4, 8, 12, 13, 16, 17, 20])
def test_grid_sampler_fulls_grid(n_trials):
    round, ges = _get_optimization_scenario(n_trials)

    tried: Set[Dict[str, Any]] = set()
    for i in range(n_trials):
        hp = ges.find_next_best_hyperparams(round).to_flat_dict()
        round.add_testing_optim_result(hp)
        flat = frozenset(hp.items())

        assert flat not in tried, f"Found the same hyperparameter samples twice: `{flat}`, with {i+1} past trials."
        tried.add(flat)
    assert len(tried) == n_trials


@pytest.mark.parametrize('n_trials', [1, 3, 4, 12, 13, 16, 17, 20, 40, 50, 100])
def test_grid_sampler_fulls_individual_params(n_trials):
    round_scope, ges = _get_optimization_scenario(n_trials=n_trials)
    ges: GridExplorationSampler = ges  # typing
    round_scope: Round = round_scope  # typing

    tried_params: Dict[str, Set[Any]] = defaultdict(set)
    for i in range(n_trials):
        hp = ges.find_next_best_hyperparams(round_scope).to_flat_dict()
        round_scope.add_testing_optim_result(hp)
        for hp_k, value_set in hp.items():
            tried_params[hp_k].add(value_set)

    for hp_k, value_set in tried_params.items():
        hp_k: str = hp_k  # typing
        value_set: Set[Any] = value_set  # typing

        if isinstance(round_scope.hp_space[hp_k], DiscreteHyperparameterDistribution):
            round_numbs = min(n_trials, len(round_scope.hp_space[hp_k].values()))
            assert len(value_set) == round_numbs
        else:
            round_numbs = min(n_trials, len(ges.flat_hp_grid_values[hp_k]))
            assert len(value_set) >= round_numbs, (
                f"value_set={value_set} has a len={len(value_set)}, but len={round_numbs} was expected.")


@dataclass
class RoundStub:
    hp_space: HyperparameterSpace = field(default_factory=HyperparameterSpace)
    _all_tried_hyperparams: List[FlatDict] = field(default_factory=list)

    def add_testing_optim_result(self, rdict: HyperparameterSamples):
        self._all_tried_hyperparams.append(rdict)

    def get_all_hyperparams(self):
        return self._all_tried_hyperparams


def _get_optimization_scenario(n_trials):
    round_scope: Round = RoundStub(hp_space=HyperparameterSpace({
        'a__add_n': Uniform(0, 4),
        'b__multiply_n': RandInt(0, 4),
        'c__Pchoice': PriorityChoice(["one", "two"]),
    }))
    ges = GridExplorationSampler(n_trials)
    return round_scope, ges


class OnlyFitAtTransformTime(NonFittableMixin, MetaStep):
    """
    This is needed for AutoML to make sure that fitting will happen in parallel, not in the same thread.
    Note that nothing is returned, the step does not transform the data. It only fits.
    """

    def _transform_data_container(self, data_container: TrainDACT, context: CX) -> PredsDACT:
        self.wrapped = self.wrapped.handle_fit(data_container, context)
        return data_container


def test_automl_can_resume_last_run_and_retrain_best_with_0_trials(tmpdir):
    dact = DACT(di=list(range(10)), eo=list(range(10, 20)))
    cx = AutoMLContext.from_context(CX(root=tmpdir))
    sleep_step = Sleep(0.001)
    n_trials = 4
    automl: ControlledAutoML = _create_automl_test_loop(
        tmpdir, sleep_step, n_trials=1, start_new_round=False, refit_best_trial=False
    )
    for _ in range(n_trials):
        copy.deepcopy(automl).handle_fit(dact, cx)

    automl_refiting_best = _create_automl_test_loop(
        tmpdir, sleep_step, n_trials=0, start_new_round=False, refit_best_trial=True)
    automl_refiting_best, preds = automl_refiting_best.handle_fit_transform(dact, cx)

    round_scope: Round = cx.with_loc(ScopedLocation.default(round_number=0)).load_agg()
    bests: List[Tuple[float, int, FlatDict]] = round_scope.summary()

    assert len(bests) == n_trials
    assert len(set(tuple(dict(hp).items()) for score, i, hp in bests)
               ) == n_trials, f"Expecting unique hyperparams for the given n_trials={n_trials} and intelligent grid sampler."

    best_score = bests[0][0]
    assert median_absolute_error(dact.eo, preds.di) == best_score


def test_automl_can_use_same_repo_in_parallel(tmpdir, use_processes=True):
    # @pytest.mark.parametrize("use_processes", [False, True])
    dact = DACT(di=list(range(10)), eo=list(range(10, 20)))
    cx = AutoMLContext.from_context(CX(root=tmpdir), repo=HyperparamsOnDiskRepository(tmpdir))
    # cx.add_scoped_logger_file_handler() TODO: use OnDiskRepo to test logging files with parallel.
    sleep_step = Sleep(0.25)
    automl: ControlledAutoML = _create_automl_test_loop(
        tmpdir, sleep_step, n_trials=1, start_new_round=False, refit_best_trial=False
    )
    n_trials = 3 * 2
    parallel_automl = ParallelQueuedFeatureUnion(
        # steps=[OnlyFitAtTransformTime(automl)],
        steps=[
            OnlyFitAtTransformTime(copy.deepcopy(automl)),
            OnlyFitAtTransformTime(copy.deepcopy(automl)),
            OnlyFitAtTransformTime(copy.deepcopy(automl))
        ],
        batch_size=int(len(dact) / 2),
        n_workers_per_step=2,
        use_processes=use_processes,
        use_savers=False,
        max_queue_size=1
    )
    parallel_automl.handle_fit_transform(dact, cx)

    # OnlyFitAtTransformTime(copy.deepcopy(automl)).handle_fit_transform(dact, cx)
    # OnlyFitAtTransformTime(copy.deepcopy(automl)).handle_fit_transform(dact, cx)
    # OnlyFitAtTransformTime(copy.deepcopy(automl)).handle_fit_transform(dact, cx)
    # OnlyFitAtTransformTime(copy.deepcopy(automl)).handle_fit_transform(dact, cx)
    # OnlyFitAtTransformTime(copy.deepcopy(automl)).handle_fit_transform(dact, cx)
    # OnlyFitAtTransformTime(copy.deepcopy(automl)).handle_fit_transform(dact, cx)

    automl_refiting_best = _create_automl_test_loop(
        tmpdir, sleep_step, n_trials=0, start_new_round=False, refit_best_trial=True)
    automl_refiting_best, preds = automl_refiting_best.handle_fit_transform(dact, cx)

    round_scope: Round = cx.with_loc(ScopedLocation.default(round_number=0)).load_agg()
    bests: List[Tuple[float, int, FlatDict]] = round_scope.summary()

    assert len(bests) == n_trials

    best_score = bests[0][0]
    assert median_absolute_error(dact.eo, preds.di) == best_score



import os
import random
from typing import List

import numpy as np
import pytest
from neuraxle.base import ExecutionContext as CX
from neuraxle.hyperparams.distributions import (Choice, DistributionMixture,
                                                HPSampledValue,
                                                HyperparameterDistribution,
                                                LogNormal, LogUniform, Normal,
                                                Quantized, RandInt, Uniform)
from neuraxle.hyperparams.space import (HyperparameterSamples,
                                        HyperparameterSpace)
from neuraxle.metaopt.auto_ml import (AutoML, BaseHyperparameterOptimizer,
                                      ControlledAutoML)
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.metaopt.data.aggregates import (MetricResults, Round, Trial,
                                              TrialSplit)
from neuraxle.metaopt.data.vanilla import (AutoMLContext,
                                           HyperparameterSamplerStub,
                                           MetricResultsDataclass,
                                           RoundDataclass, ScopedLocation,
                                           TrialDataclass, TrialSplitDataclass,
                                           VanillaHyperparamsRepository)
from neuraxle.metaopt.hyperopt.tpe import (TreeParzenEstimator,
                                           _DividedMixturesFactory,
                                           _DividedTPEPosteriors)
from neuraxle.metaopt.validation import (GridExplorationSampler,
                                         ValidationSplitter)
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import AddN
from sklearn.metrics import mean_squared_error

N_TRIALS = 40


def _avg(seq: List):
    if len(seq) == 0:
        return None
    return sum(seq) / len(seq)


distributions: List[HyperparameterDistribution] = [
    LogNormal(log2_space_mean=1.0, log2_space_std=0.5, hard_clip_min=0, hard_clip_max=6),
    Choice(choice_list=[0, 1.5, 2, 3.5, 4, 5, 6]),
    LogUniform(min_included=1.0, max_included=5.0),
    Normal(mean=3.0, std=2.0, hard_clip_min=0, hard_clip_max=6),
    Quantized(Uniform(0, 10)),
    Uniform(0, 6),
]

# @pytest.mark.parametrize("reduce_func", [min, _avg])


@pytest.mark.parametrize("add_range", distributions)
def test_tpe(tmpdir, add_range: HyperparameterDistribution, reduce_func=_avg):
    np.random.seed(42)
    random.seed(42)

    # Given
    pipeline = Pipeline([
        AddN(0.).set_hyperparams_space(HyperparameterSpace({'add': add_range})),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({'add': RandInt(1, 2)})),
    ])
    half_trials = int(N_TRIALS / 2)
    tpe: BaseHyperparameterOptimizer = TreeParzenEstimator(
        number_of_initial_random_step=half_trials,
        quantile_threshold=0.3,
        number_good_trials_max_cap=25,
        number_possible_hyperparams_candidates=100,
        use_linear_forgetting_weights=False,
        number_recent_trials_at_full_weights=25
    )
    grid: BaseHyperparameterOptimizer = GridExplorationSampler(N_TRIALS)

    grid_scores = _score_meta_optimizer(grid, pipeline, tmpdir)
    tpe_scores = _score_meta_optimizer(tpe, pipeline, tmpdir)

    mean_compared_grid_score = reduce_func(grid_scores[half_trials:])
    mean_compared_tpe_score = reduce_func(tpe_scores[half_trials:])

    assert mean_compared_tpe_score <= mean_compared_grid_score


def _score_meta_optimizer(meta_optimizer: BaseHyperparameterOptimizer, pipeline, tmpdir):
    expected_output_mult = 3.5
    optim_scores = _score_trials(
        expected_output_mult,
        pipeline,
        meta_optimizer,
        os.path.join(tmpdir, meta_optimizer.__class__.__name__)
    )
    return optim_scores


def _score_trials(
    expected_output_mult,
    pipeline,
    hyperparams_optimizer: BaseHyperparameterOptimizer,
    tmpdir: str
):
    hp_repository: VanillaHyperparamsRepository = VanillaHyperparamsRepository(str(tmpdir))
    n_epochs = 1
    auto_ml: ControlledAutoML = AutoML(
        pipeline=pipeline,
        hyperparams_optimizer=hyperparams_optimizer,
        validation_splitter=ValidationSplitter(0.5),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False, name='mse'),
        n_trials=N_TRIALS,
        refit_best_trial=True,
        epochs=n_epochs,
        hyperparams_repository=hp_repository,
        continue_loop_on_error=False
    )

    # When
    data_inputs = np.array([0, 0])
    # Fitting and optimizing a multiplier problem with a model make to add...
    expected_outputs = expected_output_mult * np.ones_like(data_inputs)
    auto_ml.fit(data_inputs=data_inputs, expected_outputs=expected_outputs)

    # Then
    cx = auto_ml.get_automl_context(CX()).with_loc(ScopedLocation.default(-1))
    trials: Round = Round.from_context(cx)
    #  hp_repository.load(ScopedLocation.default(-1)).filter(TrialStatus.SUCCESS)
    trials_validation_scores: List[float] = [t.get_avg_validation_score() for t in trials]
    return trials_validation_scores


def test_divided_posteriors_rvs_good_is_in_range():
    good_trials = RandInt(1, 2)
    bad_trials = RandInt(101, 102)
    divider: _DividedTPEPosteriors = _DividedTPEPosteriors(good_trials, bad_trials)

    goods: List[int] = []
    for i in range(100):
        goods.append(divider.rvs_good())

    assert good_trials.min() <= min(goods)
    assert max(goods) <= good_trials.max()


def test_divided_posteriors_ratio_is_ok():
    good_trials = Normal(mean=1, std=0.5, hard_clip_min=0, hard_clip_max=3)
    bad_trials = Normal(mean=2, std=0.5, hard_clip_min=0, hard_clip_max=3)
    divider: _DividedTPEPosteriors = _DividedTPEPosteriors(good_trials, bad_trials)
    gmm: DistributionMixture = DistributionMixture.build_gaussian_mixture(
        distribution_amplitudes=[0.5, 0.5],
        means=[1, 2],
        stds=[0.5, 0.5],
        distributions_mins=[0, 0],
        distributions_max=[3, 3],
    )

    for i in range(100):
        good_hyperparam, proba_ratio = divider.rvs_good_with_pdf_division_proba()

        assert proba_ratio > 0.0
        if good_hyperparam > gmm.mean():
            assert proba_ratio < 1.0
        else:
            assert proba_ratio >= 1.0
    assert gmm.mean() == 1.5


class TrialBuilder:

    def __init__(self, round_scope: Round) -> None:
        self.round_scope: Round = round_scope

    def with_optimizer(
        self,
        hp_optimizer: BaseHyperparameterOptimizer = None,
        hp_space: HyperparameterSpace = None,
    ) -> 'TrialBuilder':
        self.round_scope.with_optimizer(hp_optimizer, hp_space)
        self.round_scope.save()
        return self

    def add_trial_from_space_rvs(
        self,
        score: float,
        metric_name: str,
        is_higher_score_better: bool,
        hyperparams_samples: HyperparameterSamples = None,
    ) -> 'TrialBuilder':

        if hyperparams_samples is not None:
            self.round_scope.hp_optimizer = HyperparameterSamplerStub(hyperparams_samples)

        with self.round_scope.new_rvs_trial() as trial:
            trial: Trial = trial

            with trial.new_validation_split() as trial_split:
                trial_split: TrialSplit = trial_split

                with trial_split.managed_metric(metric_name, is_higher_score_better) as metric:
                    metric: MetricResults = metric

                    metric.add_train_result(score)
                    metric.add_valid_result(score)

        self.round_scope.save()
        return self

    def add_many_trials_from_space_rvs(
        self,
        scores: List[float],
        metric_name: str,
        is_higher_score_better: bool,
        hp_samples: List[HyperparameterSamples] = None,
    ) -> 'TrialBuilder':
        if hp_samples is None:
            hp_samples = [None] * len(scores)
        for score, hp in zip(scores, hp_samples):
            self.add_trial_from_space_rvs(score, metric_name, is_higher_score_better, hp)
        return self

    def build(self) -> Round:
        return self.round_scope.save()


def test_divided_mixtures_factory():
    hp_space = HyperparameterSpace([('a', Uniform(0, 1)), ('b', LogUniform(1, 2))])
    round_scope: Round = Round.dummy()
    round_scope.with_optimizer(None, hp_space).with_metric('mse')
    round_scope = TrialBuilder(round_scope).add_many_trials_from_space_rvs(
        scores=[0.0, 0.1, 0.9, 1.0],
        metric_name='mse',
        is_higher_score_better=False,
        hp_samples=[
            HyperparameterSamples([('a', 0.1), ('b', 1.1)]),
            HyperparameterSamples([('a', 0.9), ('b', 1.1)]),
            HyperparameterSamples([('a', 0.1), ('b', 1.9)]),
            HyperparameterSamples([('a', 0.9), ('b', 1.9)]),
        ]
    ).build()

    dmf: _DividedMixturesFactory = _DividedMixturesFactory(
        quantile_threshold=0.5,
        number_good_trials_max_cap=10,
        use_linear_forgetting_weights=False,
        number_recent_trials_at_full_weights=10
    )
    hps_names = List[str]
    divided_distributions = List['_DividedTPEPosteriors']

    hps_names, divided_distributions = dmf.create_from(round_scope)
    a_good_trials = divided_distributions[0].good_trials
    a_bad_trials = divided_distributions[0].bad_trials
    b_good_trials = divided_distributions[1].good_trials
    b_bad_trials = divided_distributions[1].bad_trials

    assert a_good_trials.mean() == 0.5
    assert a_bad_trials.mean() == 0.5
    assert b_good_trials.mean() < 1.5
    assert b_bad_trials.mean() > 1.5

    best_ratios = []
    best_samples = []
    for i, dist in enumerate(divided_distributions):
        best_ratio = 0
        best_hp_sample = None
        for _ in range(100):

            hp_sample, ratio = dist.rvs_good_with_pdf_division_proba()
            if ratio > best_ratio or best_hp_sample is None:
                best_ratio = ratio
                best_hp_sample = hp_sample

        best_ratios.append(best_ratio)
        best_samples.append(best_hp_sample)

    assert best_ratios[0] == 1.0
    assert best_ratios[1] > 1.0
    assert 1.0 <= best_samples[1] < 1.2


@pytest.mark.parametrize("dist", [
    LogNormal(log2_space_mean=1.0, log2_space_std=0.5, hard_clip_min=0, hard_clip_max=6),
    LogNormal(log2_space_mean=1.0, log2_space_std=0.5, hard_clip_min=0),
    Choice(choice_list=[0, 1, 2, 3, 4, 5, 6]),
    LogUniform(min_included=1.0, max_included=5.0),
    Normal(mean=3.0, std=2.0, hard_clip_min=0, hard_clip_max=6),
    Normal(mean=3.0, std=2.0),
    Quantized(Uniform(0, 10)),
    Uniform(0, 6),
])
def test_icdf_is_within_tolerance_value(dist: HyperparameterDistribution):
    tolerance = 2e-5

    # if isinstance(dist, ContinuousHyperparameterDistribution):
    #     dist.min = dist._pseudo_min
    #     dist.max = dist._pseudo_max
    _min = dist.min() if dist.is_discrete() else dist._pseudo_min()
    _max = dist.max() if dist.is_discrete() else dist._pseudo_max()

    # create x range
    if dist.is_discrete():
        x_range = dist.values()
    else:
        x_range = np.linspace(_min, _max, 15).tolist()[1:-1]
    # recreate p from x range
    p_range = [dist.cdf(i) for i in x_range]

    # expand p and x with more points
    _ps: List[float] = [dist.cdf(_min), dist.cdf(_max)] + p_range + [-0.1, 1.1]
    _xs: List[HPSampledValue] = [_min, _max] + x_range + [_min, _max]

    # inverse with icdf to find x again and assert x_icdf == x
    for p, x in zip(_ps, _xs):
        x_icdf = dist.icdf(p)
        assert abs(x_icdf - x) <= tolerance, f"`x_icdf != x` : `{x_icdf} != {x}`, given `p={p}`"



import pytest
from py._path.local import LocalPath


def test_auto_ml_loop_clean_kata(tmpdir: LocalPath):
    from examples.auto_ml.plot_automl_loop_clean_kata import main
    main(tmpdir)


def test_easy_rest_api_serving():
    from examples.deployment.plot_easy_rest_api_serving import main
    main()


def test_force_handle_mixin():
    from examples.getting_started.plot_force_handle_mixin import main
    main()


def test_inverse_transform():
    from examples.getting_started.plot_inverse_transform import main
    main()


def test_label_encoder_across_multiple_columns():
    from examples.getting_started.plot_label_encoder_across_multiple_columns import main
    main()


def test_nested_pipelines():
    from examples.getting_started.plot_nested_pipelines import main
    main()


def test_non_fittable_mixin():
    from examples.getting_started.plot_non_fittable_mixin import main
    main()


def test_hyperparams():
    from examples.hyperparams.plot_hyperparams import main
    main()


def test_apply():
    from examples.operations.plot_apply_method import main
    main()


def test_parallel_streaming():
    from examples.parallel.plot_streaming_pipeline import main
    main()


def test_boston_housing_meta_optimization(tmpdir: LocalPath):
    from examples.sklearn.plot_boston_housing_meta_optimization import main
    main(tmpdir)


def test_boston_housing_regression_with_model_stacking():
    from examples.sklearn.plot_boston_housing_regression_with_model_stacking import main
    main()


def test_cyclical_feature_engineering():
    from examples.sklearn.plot_cyclical_feature_engineering import predictions
    print(predictions)
    assert predictions is not None



import math
import os
from collections import Counter

import joblib
import numpy as np
import pytest
from neuraxle.base import Identity
from neuraxle.hyperparams.distributions import (Choice, LogNormal, LogUniform,
                                                Normal, PriorityChoice,
                                                RandInt, Uniform,
                                                get_index_in_list_with_bool)
from neuraxle.hyperparams.scipy_distributions import (
    Gaussian, Histogram, Poisson, ScipyContinuousDistributionWrapper,
    ScipyDiscreteDistributionWrapper, ScipyLogUniform, StdMeanLogNormal)
from neuraxle.hyperparams.space import HyperparameterSpace
from scipy.stats import gamma, norm, randint, uniform

NUM_TRIALS = 100


def get_many_samples_for(hd):
    return [hd.rvs() for _ in range(NUM_TRIALS)]


def test_wrapped_sk_learn_distributions_should_be_able_to_use_sklearn_methods():
    wrapped_sklearn_distribution = Gaussian(min_included=0, max_included=10, null_default_value=0)

    assert wrapped_sklearn_distribution.logpdf(5) == -13.418938533204672
    assert wrapped_sklearn_distribution.logcdf(5) == -0.6931477538632531
    assert wrapped_sklearn_distribution.sf(5) == 0.5000002866515718
    assert wrapped_sklearn_distribution.logsf(5) == -0.693146607256966
    assert np.all(wrapped_sklearn_distribution.ppf([0.0, 0.01, 0.05, 0.1, 1 - 0.10, 1 - 0.05, 1 - 0.01, 1.0], 10))
    assert 8 < wrapped_sklearn_distribution.isf(q=0.5) > 8
    assert wrapped_sklearn_distribution.moment(2) > 50
    stats = wrapped_sklearn_distribution.stats()
    assert stats[0]
    assert stats[1]
    assert np.array_equal(wrapped_sklearn_distribution.entropy(), np.array(0.7094692666023363))
    assert wrapped_sklearn_distribution.median()
    assert wrapped_sklearn_distribution.mean() == 5.398942280397029
    assert np.isclose(wrapped_sklearn_distribution.std(), 4.620759921685375)
    assert np.isclose(wrapped_sklearn_distribution.var(), 21.351422253853833)
    assert wrapped_sklearn_distribution.expect() == 0.39894228040143276
    interval = wrapped_sklearn_distribution.interval(alpha=[0.25, 0.50])
    assert np.all(interval[0])
    assert np.all(interval[1])
    assert wrapped_sklearn_distribution.support() == (0, 10)


HIST_DATA = norm.rvs(size=1000, loc=0, scale=1.5, random_state=123)


def test_histogram():
    hist_dist = Histogram(
        histogram=np.histogram(norm.rvs(size=1000, loc=0, scale=1.5, random_state=123), bins=10),
        null_default_value=0.0
    )

    _test_histogram(hist_dist)


def _test_histogram(hist_dist: Histogram):
    assert min(HIST_DATA) <= hist_dist.rvs() <= max(HIST_DATA)
    assert 1.0 > hist_dist.pdf(x=1.0) > 0.0
    assert hist_dist.pdf(x=np.max(HIST_DATA)) == 0.0
    assert hist_dist.pdf(x=np.min(HIST_DATA)) < 0.05
    assert hist_dist.cdf(x=np.max(HIST_DATA)) == 1.0
    assert 0.55 > hist_dist.cdf(x=np.median(HIST_DATA)) > 0.45
    assert hist_dist.cdf(x=np.min(HIST_DATA)) == 0.0


def test_continuous_gaussian():
    gaussian_distribution = Gaussian(
        min_included=0,
        max_included=10,
        null_default_value=0.0
    )

    _test_gaussian(gaussian_distribution)


def _test_gaussian(gaussian_distribution):
    assert 0.0 <= gaussian_distribution.rvs() <= 10.0
    assert gaussian_distribution.pdf(10) < 0.001
    assert gaussian_distribution.pdf(0) < 0.42
    assert 0.55 > gaussian_distribution.cdf(5.0) > 0.45
    assert gaussian_distribution.cdf(0) == 0.0


def test_discrete_poison():
    poisson_distribution = Poisson(
        min_included=0.0,
        max_included=10.0,
        null_default_value=0.0,
        mu=5.0
    )

    _test_discrete_poisson(poisson_distribution)


def _test_discrete_poisson(poisson_distribution: Poisson):
    rvs = [poisson_distribution.rvs() for i in range(10)]
    assert not all(x == rvs[0] for x in rvs)
    assert 0.0 <= poisson_distribution.rvs() <= 10.0
    assert poisson_distribution.pdf(10) == 0.01813278870782187
    assert np.isclose(poisson_distribution.pdf(0), 0.006737946999085467)
    assert poisson_distribution.cdf(5.0) == 0.6159606548330632
    assert poisson_distribution.cdf(0) == 0.006737946999085467


def test_randint():
    hd = RandInt(min_included=-10, max_included=10, null_default_value=0)

    _test_randint(hd)


def _test_randint(hd):
    samples = hd.rvs_many(size=100)
    samples_mean = np.abs(np.mean(samples))
    invprob = 1 / (10 + 10 + 1)
    assert -5.0 < samples_mean < 5.0
    assert min(samples) >= -10.0
    assert max(samples) <= 10.0
    assert hd.pdf(-11) == 0.
    assert hd.pdf(-10) == invprob
    assert hd.pdf(0) == invprob
    assert hd.pdf(5) == invprob
    assert hd.pdf(10) == invprob
    assert hd.pdf(11) == 0.
    assert hd.cdf(-10.1) == 0.
    assert hd.cdf(-10) == invprob
    assert hd.cdf(5) == 16 * invprob
    assert abs(hd.cdf(10) - 1.) < 1e-6
    assert hd.cdf(10.1) == 1.
    assert hd.min() == -10
    assert hd.mean() == 0
    assert hd.std() > 2
    assert hd.max() == 10


def test_uniform():
    hd = Uniform(-10, 10)
    _test_uniform(hd)


def _test_uniform(hd):
    samples = hd.rvs_many(size=100)
    samples_mean = np.abs(np.mean(samples))

    assert samples_mean < 4.0
    assert min(samples) >= -10.0
    assert max(samples) <= 10.0
    assert hd.pdf(-10.1) == 0.
    assert abs(hd.pdf(0) - 1 / (10 + 10)) < 1e-6
    assert hd.pdf(10.1) == 0.
    assert hd.cdf(-10.1) == 0.
    assert abs(hd.cdf(0) - (0 + 10) / (10 + 10)) < 1e-6
    assert hd.cdf(10.1) == 1.


def test_loguniform():
    hd = ScipyLogUniform(min_included=0.001, max_included=10)

    _test_loguniform(hd)


def _test_loguniform(hd: ScipyLogUniform):
    samples = hd.rvs_many(size=100)
    samples_mean = np.mean(samples)
    assert samples_mean < 1.5
    assert min(samples) >= 0.001
    assert max(samples) <= 10.0
    assert hd.pdf(0.0001) == 0.
    assert abs(hd.pdf(2) - 0.054286810237906484) < 2e-6
    assert hd.pdf(10.1) == 0.
    assert hd.cdf(0.0001) == 0.
    assert abs(hd.cdf(2) - (math.log2(2) - math.log2(0.001)) / (math.log2(10) - math.log2(0.001))) < 1e-6
    assert hd.cdf(10.1) == 1.


def test_normal():
    hd = Normal(
        hard_clip_min=0.0,
        hard_clip_max=1.0,
        mean=0.5,
        std=0.2,
        null_default_value=0.0
    )

    _test_normal(hd)


def _test_normal(hd):
    samples = hd.rvs_many(size=100)
    samples_mean = np.abs(np.mean(samples))
    assert 0.6 > samples_mean > 0.4
    samples_std = np.std(samples)
    assert 0.1 < samples_std < 0.6
    assert abs(hd.pdf(-1.) - 0.24) == 0.24
    assert abs(hd.pdf(0.) - 0.40) == 0.31125636093539194
    assert abs(hd.pdf(1.)) == 0.08874363906460808
    assert abs(hd.cdf(-1.) - 0.15) == 0.15
    assert abs(hd.cdf(0.) - 0.5) == 0.5
    assert abs(hd.cdf(1.) - 0.85) == 0.15000000000000002


def test_lognormal():
    hd = StdMeanLogNormal(
        hard_clip_min=-5,
        hard_clip_max=5,
        log2_space_mean=0.0,
        log2_space_std=2.0,
        null_default_value=-1.0
    )

    _test_lognormal(hd)


def _test_lognormal(hd: StdMeanLogNormal):
    samples = hd.rvs_many(size=100)
    samples_median = np.median(samples)
    assert -5 < samples_median < 5
    samples_std = np.std(samples)
    assert 0 < samples_std < 4
    assert hd.pdf(0.) == 0.
    assert abs(hd.pdf(1.) - 0.28777602476804065) < 1e-6
    assert abs(hd.pdf(5.) - 0.029336304593386688) < 1e-6
    assert hd.cdf(0.) == 0.
    assert hd.cdf(1.) == 0.49999999998280026
    assert abs(hd.cdf(5.) - 0.8771717397015799) == 0.12282826029842009


@pytest.mark.parametrize("hd, test_method", [
    (RandInt(min_included=-10, max_included=10, null_default_value=0), _test_randint),
    (StdMeanLogNormal(hard_clip_min=-5, hard_clip_max=5, log2_space_mean=0.0,
     log2_space_std=2.0, null_default_value=-1.0), _test_lognormal),
    (Normal(hard_clip_min=0.0, hard_clip_max=1.0, mean=0.5, std=0.2, null_default_value=0.0), _test_normal),
    (ScipyLogUniform(min_included=0.001, max_included=10), _test_loguniform),
    (Uniform(min_included=-10, max_included=10), _test_uniform),
    (Poisson(min_included=0.0, max_included=10.0, null_default_value=0.0, mu=5.0), _test_discrete_poisson),
    (Gaussian(min_included=0, max_included=10, null_default_value=0.0), _test_gaussian),
    (Histogram(histogram=np.histogram(HIST_DATA, bins=10), null_default_value=0.0), _test_histogram)
])
def test_after_serialization(hd, test_method, tmpdir):
    joblib.dump(hd, os.path.join(str(tmpdir), '{}.joblib'.format(hd.__class__.__name__)))
    hd_loaded = joblib.load(os.path.join(str(tmpdir), '{}.joblib'.format(hd.__class__.__name__)))

    assert hd.__class__ == hd_loaded.__class__
    test_method(hd_loaded)


@pytest.mark.parametrize("hd", [
    Poisson(min_included=0.0, max_included=10.0, null_default_value=0.0, mu=5.0),
    Choice(choice_list=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
    PriorityChoice(choice_list=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
])
def test_discrete_probabilities(hd):
    probas = hd.probabilities()

    sum_probas = sum(probas)
    assert sum_probas > 0.98
    assert len(probas) == 11


def test_can_set_scipy_distribution():
    space = HyperparameterSpace({
        'rand_int_scipy': randint(low=2, high=5),  # scipy
        'rand_int_neuraxle': RandInt(2, 5),  # neuraxle
        'gamma_scipy': gamma(0.2),  # scipy
    })

    p = Identity().set_hyperparams_space(space)

    rand_int_scipy = p.get_hyperparams_space()['rand_int_scipy']
    assert isinstance(rand_int_scipy, ScipyDiscreteDistributionWrapper)
    for _ in range(20):
        randint_sample = rand_int_scipy.rvs()
        assert randint_sample in rand_int_scipy

    gamma_scipy = p.get_hyperparams_space()['gamma_scipy']
    assert isinstance(gamma_scipy, ScipyContinuousDistributionWrapper)
    for _ in range(20):
        gamma_sample = gamma_scipy.rvs()
        assert isinstance(gamma_sample, float)
        assert gamma_sample in gamma_scipy


def test_can_update_scipy_distribution():
    p = Identity().set_hyperparams_space(HyperparameterSpace({
        'rand_int_neuraxle': RandInt(2, 5)  # neuraxle
    }))

    p.update_hyperparams_space(HyperparameterSpace({
        'rand_int_scipy': randint(low=2, high=5),  # scipy
        'gamma_scipy': gamma(0.2),  # scipy
    }))

    assert isinstance(p.get_hyperparams_space()['rand_int_scipy'], ScipyDiscreteDistributionWrapper)
    assert isinstance(p.get_hyperparams_space()['gamma_scipy'], ScipyContinuousDistributionWrapper)
    randint_sample = p.get_hyperparams_space()['rand_int_scipy'].rvs()
    gamma_sample = p.get_hyperparams_space()['gamma_scipy'].rvs()
    assert 5 >= randint_sample >= 2
    assert isinstance(gamma_sample, float)
